"""
Copyright (C) 2022 - Mahesh Patil <mbpatil@ee.iitb.ac.in>
This file is part of GSEIM.

GSEIM is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

class SolveBlock():
    def __init__(self, d_slvparms={}, s_name='', s_index='', d_solve_block={}):
        self.d_parms = {}
        self.l_out = []

#       Need to take care of two situations:
#       1. The solve block data is read from a file (a saved grc file)
#          and passed as d_solve_block.
#       2. The solve block default parms are read from a file and passed
#          as d_slvparms. In this case, we also expect s_name and s_index
#          to be specified.
#
#       Notes:
#       - Always pass d_slvparms when calling SolveBlock.
#       - even if a stored solve block is being treated, first
#         assign the default values and then write over those

        for k, v in d_slvparms.items():
            self.d_parms[k] = v['default']

        if s_name:
            self.name = s_name
        if s_index:
            self.index = s_index
            self.d_parms['block_index'] = s_index

        if len(d_solve_block) != 0:
            self.name = d_solve_block['name']
            self.index = d_solve_block['index']

#           Note: no need to assign self.d_parms['block_index']; it will be
#             assigned from d_parms anyway

            for out in d_solve_block['output_blocks']:
                self.l_out.append(out)

            for k in d_solve_block['d_parms'].keys():
                if k in d_slvparms.keys():
                    self.d_parms[k] = d_solve_block['d_parms'][k]

    def block_to_dict(self):
        d = {}
        d['name'] = self.name
        d['index'] = self.index
        d['output_blocks'] = self.l_out
        d['d_parms'] = self.d_parms
        return d



"""
Copyright (C) 2022 - Mahesh Patil <mbpatil@ee.iitb.ac.in>
This file is part of GSEIM.

GSEIM is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import gi
gi.require_version("Gtk", "3.0")
from gi.repository import Gtk, Pango

class ShowList(Gtk.Window):
    def __init__(self, l, W, H, s_title):
        Gtk.Window.__init__(self, title=s_title)

#       self.set_default_size(-1, 350)
        self.set_default_size(W, H)
        self.set_position(Gtk.WindowPosition.CENTER)

        self.grid = Gtk.Grid()
        self.add(self.grid)

        self.create_textview(l)

    def create_textview(self, l):
        scrolledwindow = Gtk.ScrolledWindow()
        scrolledwindow.set_hexpand(True)
        scrolledwindow.set_vexpand(True)
        self.grid.attach(scrolledwindow, 0, 1, 3, 1)

        self.textview = Gtk.TextView()
        self.textview.override_font(
          Pango.font_description_from_string('DejaVu Sans Mono 10')
        )
        self.textbuffer = self.textview.get_buffer()

        text1 = ''
        for l1 in l:
            text1 += l1[0] + l1[1] + '\n'

        self.textbuffer.set_text(text1)

        scrolledwindow.add(self.textview)



from .utils import Spec, expand, str_

PARAM_SCHEME = expand(
    base_key=str_,   # todo: rename/remove

    id=str_,
    label=str_,
    category=str_,

    dtype=str_,
    default=object,

    options=list,
    option_labels=list,
    option_attributes=Spec(types=dict, required=False, item_scheme=(str_, list)),

    hide=str_,
)
#PORT_SCHEME = expand(
#    label=str_,
#    domain=str_,
#
#    id=str_,
#    dtype=str_,
#    vlen=(int, str_),
#
#    multiplicity=(int, str_),
#    optional=(bool, int, str_),
#    hide=(bool, str_),
#)
PORT_SCHEME = expand(
    label=str_,
    domain=str_,

    id=str_,
    dtype=str_,

    hide=(bool, str_),
)
OUTVAR_SCHEME = expand(
    id=str_,
    outvars=list,
)

BLOCK_SCHEME = expand(
    id=Spec(types=str_, required=True, item_scheme=None),
    label=str_,
    category=str_,
    flags=(list, str_),

    parameters=Spec(types=list, required=False, item_scheme=PARAM_SCHEME),
    inputs=Spec(types=list, required=False, item_scheme=PORT_SCHEME),
    outputs=Spec(types=list, required=False, item_scheme=PORT_SCHEME),

    e_left_nodes=Spec(types=list, required=False, item_scheme=PORT_SCHEME),
    e_right_nodes=Spec(types=list, required=False, item_scheme=PORT_SCHEME),
    e_top_nodes=Spec(types=list, required=False, item_scheme=PORT_SCHEME),
    e_bottom_nodes=Spec(types=list, required=False, item_scheme=PORT_SCHEME),

    b_left_nodes=Spec(types=list, required=False, item_scheme=PORT_SCHEME),
    b_right_nodes=Spec(types=list, required=False, item_scheme=PORT_SCHEME),
    b_top_nodes=Spec(types=list, required=False, item_scheme=PORT_SCHEME),
    b_bottom_nodes=Spec(types=list, required=False, item_scheme=PORT_SCHEME),

    outvars=(list, str_),

    asserts=(list, str_),
    value=str_,

    documentation=str_,
    grc_source=str_,

    block_wrapper_path=str_,  # todo: rename/remove
)



"""
Copyright (C) 2022 - Mahesh Patil <mbpatil@ee.iitb.ac.in>
This file is part of GSEIM.

GSEIM is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import sys

def east_of(p1, p2):
    if p1[0] > p2[0]:
        return p1
    else:
        return p2

def west_of(p1, p2):
    if p1[0] < p2[0]:
        return p1
    else:
        return p2

def north_of(p1, p2):
    if p1[1] < p2[1]:
        return p1
    else:
        return p2

def south_of(p1, p2):
    if p1[1] > p2[1]:
        return p1
    else:
        return p2

def block_rel_pos_x(b1,b2):

#   b1, b2 have the form ((x,y), L, W)
#   (x,y) is the top left position of the block (including the margins).
#   L and W are the horizontal and vertical dimensions, respectively.
#   Note: y-axis direction in the GUI is opp of the normal direction
#     we will follow the same convention here.
#   dely_12: positive if the S edge of b1 is above the N edge of b2.
#   delx_12: positive if the W edge of b1 is on the right of the E edge of b2.

    b1_L = b1[1]
    b2_L = b2[1]

    x_b1_west  = b1[0][0]
    x_b1_east  = x_b1_west + b1_L

    x_b2_west  = b2[0][0]
    x_b2_east  = x_b2_west + b2_L

    delx_b12 = x_b1_west - x_b2_east
    delx_b21 = x_b2_west - x_b1_east

    return delx_b12, delx_b21

def block_rel_pos_y(b1,b2):

    b1_W = b1[2]
    b2_W = b2[2]

    y_b1_north = b1[0][1]
    y_b1_south = y_b1_north + b1_W

    y_b2_north = b2[0][1]
    y_b2_south = y_b2_north + b2_W

    dely_b12 = y_b2_north - y_b1_south
    dely_b21 = y_b1_north - y_b2_south

    return dely_b12, dely_b21

def make_block_1(p0, L0, W0, port_width, delta1):

#   p0 = (x0,y0): top left corner of gseim block
#   L0, W0: horizontal and vertical dimensions of the gseim block
#     (AFTER rotation, if any)
#   delta1: min length of connector starting from the port in the same
#     direction as the port
#   return block in the form ((x,y), L, W)

    delta = port_width + delta1
    x0p = p0[0] - delta
    y0p = p0[1] - delta
    L0p = L0 + 2*delta
    W0p = W0 + 2*delta

    return ((x0p,y0p), L0p, W0p)

def get_wire(b1, cc1, dir1, draw_arrow1, b2, cc2, dir2, draw_arrow2,
    delta1, arrow_h, arrow_w):

    flag_flip = False

    if dir1 in ('N', 'S') and dir2 in ('E', 'W'):
        flag_flip = True
    elif dir1 == 'E' and dir2 == 'W':
        flag_flip = True
    elif dir1 == 'N' and dir2 == 'S':
        flag_flip = True
    elif dir1 == 'E' and dir2 == 'E':
        if west_of(cc1, cc2) == cc2:
            flag_flip = True
    elif dir1 == 'W' and dir2 == 'W':
        if east_of(cc1, cc2) == cc2:
            flag_flip = True
    elif dir1 == 'N' and dir2 == 'N':
        if south_of(cc1, cc2) == cc2:
            flag_flip = True
    elif dir1 == 'S' and dir2 == 'S':
        if north_of(cc1, cc2) == cc2:
            flag_flip = True

    if draw_arrow1 and draw_arrow2:
        print('get_wire: draw_arrow1 and draw_arrow2 cannot both be True. Halting...')
        sys.exit()

    if flag_flip:
        block1, c1, direction1, arrow1 = (b2, cc2, dir2, draw_arrow2)
        block2, c2, direction2, arrow2 = (b1, cc1, dir1, draw_arrow1)
    else:
        block1, c1, direction1, arrow1 = (b1, cc1, dir1, draw_arrow1)
        block2, c2, direction2, arrow2 = (b2, cc2, dir2, draw_arrow2)

    xc1, yc1 = c1
    xc2, yc2 = c2

    block1_N = block1[0][1]
    block1_S = block1_N + block1[2]
    block1_W = block1[0][0]
    block1_E = block1_W + block1[1]

    block2_N = block2[0][1]
    block2_S = block2_N + block2[2]
    block2_W = block2[0][0]
    block2_E = block2_W + block2[1]

    if direction1 == 'E':
        x1p, y1p = c1p = (xc1+delta1, yc1)
    elif direction1 == 'W':
        x1p, y1p = c1p = (xc1-delta1, yc1)
    elif direction1 == 'S':
        x1p, y1p = c1p = (xc1, yc1+delta1)
    elif direction1 == 'N':
        x1p, y1p = c1p = (xc1, yc1-delta1)

    if direction2 == 'E':
        x2p, y2p = c2p = (xc2+delta1, yc2)
    elif direction2 == 'W':
        x2p, y2p = c2p = (xc2-delta1, yc2)
    elif direction2 == 'S':
        x2p, y2p = c2p = (xc2, yc2+delta1)
    elif direction2 == 'N':
        x2p, y2p = c2p = (xc2, yc2-delta1)

    l_middle = []

#   split the connection into three parts: l_begin, l_middle, l_end
#   l_begin, l_end are independent of the block locations

    dely_b12, dely_b21 = block_rel_pos_y(block1,block2)
    delx_b12, delx_b21 = block_rel_pos_x(block1,block2)

    if direction1 == 'E':
        if direction2 == 'S':
            if east_of(c1p, c2p) == c2p:
                if north_of(c1p, c2p) == c2p: # NE b1E b2S
                    l_middle.append((x2p, y1p))
                else: # SE b1E b2S
                    if dely_b12 > 0 or delx_b21 > 0:
                        if delx_b21 > 0:
                            xp = block1_E + int(delx_b21/2)
                            l_middle.append((xp, y1p))
                            l_middle.append((xp, y2p))
                        else:
                            yp = block1_S + int(dely_b12/2)
                            l_middle.append((x1p, yp))
                            l_middle.append((block2_W, yp))
                            l_middle.append((block2_W, y2p))
            else:
                if north_of(c1p, c2p) == c2p: # NW b1E b2S
                    if dely_b21 > 0:
                        yp = block2_S + int(dely_b21/2)
                        l_middle.append((x1p, yp))
                        l_middle.append((x2p, yp))
                    else:
                        if delx_b12 > 0:
                            yp = max(block1_S, block2_S)
                            l_middle.append((x1p, yp))
                            l_middle.append((x2p, yp))
                else: # SW b1E b2S
                    if dely_b12 > 0:
                        if block1_E > block2_E:
                            if block2_S > block1_S:
                                l_middle.append((x1p, block2_S))
                            else:
                                l_middle.append((x1p, block1_S))
                                l_middle.append((x2p, block1_S))
                        else:
                            yp = block1_S + int(dely_b12/2)
                            l_middle.append((x1p, yp))
                            l_middle.append((block2_E, yp))
                            l_middle.append((block2_E, y2p))
                    else:
                        if delx_b12 > 0:
                            if block1_S > block2_S:
                                l_middle.append((x1p, block1_S))
                                l_middle.append((x2p, block1_S))
                            else:
                                l_middle.append((x1p, y2p))
        elif direction2 == 'N':
            if east_of(c1p, c2p) == c2p:
                if south_of(c1p, c2p) == c2p: # SE b1E b2N
                    if delx_b21 > 0:
                        l_middle.append((x2p, y1p))
                    elif dely_b12 > 0:
                        l_middle.append((x2p, y1p))
                else: # NE b1E b2N
                    if delx_b21 > 0:
                        xp = block1_E + int(delx_b21/2)
                        l_middle.append((xp, y1p))
                        l_middle.append((xp, y2p))
                    elif dely_b21 > 0:
                        l_middle.append((block2_E, y1p))
                        l_middle.append((block2_E, y2p))
            else:
                if south_of(c1p, c2p) == c2p: # SW b1E b2N
                    if delx_b12 > 0 or dely_b12 > 0:
                        if dely_b12 > 0:
                            yp = block1_S + int(dely_b12/2)
                            l_middle.append((x1p, yp))
                            l_middle.append((x2p, yp))
                        else:
                            yp = min(block1_N, block2_N)
                            l_middle.append((x1p, yp))
                            l_middle.append((x2p, yp))
                else: # NW b1E b2N
                    if delx_b12 > 0 or dely_b21 > 0:
                        if block2_N > block1_N:
                            if block2_E > block1_E:
                                l_middle.append((block2_E, y1p))
                                l_middle.append((block2_E, y2p))
                            else:
                                l_middle.append((x1p, block1_N))
                                l_middle.append((x2p, block1_N))
                        else:
                            if block2_E > block1_E:
                                l_middle.append((block2_E, y1p))
                                l_middle.append((block2_E, y2p))
                            else:
                                l_middle.append((x1p, y2p))
        elif direction2 == 'E': # b1E b2E
            if dely_b12 > 0 or dely_b21 > 0 or delx_b12 > 0 or delx_b21 > 0:
                if block2_S < y1p or block2_N > y1p:
                    l_middle.append((x2p, y1p))
                else:
                    if block2_W < x1p:
                        l_middle.append((x2p, y1p))
                    else:
                        delx = block2_W - x1p
                        xp = x1p + int(delx/2)
                        l_middle.append((xp, y1p))
                        l_middle.append((xp, block2_S))
                        l_middle.append((x2p, block2_S))
    elif direction1 == 'W':
        if direction2 == 'S':
            if delx_b12 > 0 or delx_b21 > 0 or dely_b12 > 0 or dely_b21 > 0:
                if west_of(c1p, c2p) == c2p:
                    if north_of(c1p, c2p) == c2p: # NW b1W b2S
                        l_middle.append((x2p, y1p))
                    else: # SW b1W b2S
                        if delx_b12 > 0:
                            xp = block2_E + int(delx_b12/2)
                            l_middle.append((xp, y1p))
                            l_middle.append((xp, y2p))
                        else:
                            if block2_N > y1p:
                                l_middle.append((block2_W, y1p))
                                l_middle.append((block2_W, y2p))
                            else:
                                l_middle.append((x1p, y2p))
                else:
                    if north_of(c1p, c2p) == c2p: # NE b1W b2S
                        if dely_b21 > 0:
                            yp = block2_S + int(dely_b21/2)
                            l_middle.append((x1p, yp))
                            l_middle.append((x2p, yp))
                        else:
                            if delx_b21 > 0:
                                xp = block1_E + int(delx_b21/2)
                                l_middle.append((x1p, block1_N))
                                l_middle.append((xp, block1_N))
                                l_middle.append((xp, y2p))
                    else: # SE b1W b2S
                        if block2_W < block1_W:
                            if block2_S > block1_S:
                                yp = block1_S + int(dely_b12/2)
                                l_middle.append((x1p, yp))
                                l_middle.append((block2_W, yp))
                                l_middle.append((block2_W, y2p))
                        else:
                            if block2_S > block1_S:
                                l_middle.append((x1p, y2p))
                            else:
                                l_middle.append((x1p, block1_S))
                                l_middle.append((x2p, block1_S))
        elif direction2 == 'N':
            if delx_b12 > 0 or delx_b21 > 0 or dely_b12 > 0 or dely_b21 > 0:
                if west_of(c1p, c2p) == c2p:
                    if south_of(c1p, c2p) == c2p: # SW b1W b2N
                        l_middle.append((x2p, y1p))
                    else: # NW b1W b2N
                        if delx_b12 > 0:
                            xp = block2_E + int(delx_b12/2)
                            l_middle.append((xp, y1p))
                            l_middle.append((xp, y2p))
                        else:
                            l_middle.append((block2_W, y1p))
                            l_middle.append((block2_W, y2p))
                else:
                    if south_of(c1p, c2p) == c2p: # SE b1W b2N
                        if dely_b12 > 0:
                            yp = block1_S + int(dely_b12/2)
                            l_middle.append((x1p, yp))
                            l_middle.append((x2p, yp))
                        else:
                            if delx_b21 > 0:
                                xp = block1_E + int(delx_b21/2)
                                l_middle.append((x1p, block1_S))
                                l_middle.append((xp, block1_S))
                                l_middle.append((xp, y2p))
                            else:
                                yp = max(block1_S, block2_S)
                                l_middle.append((x1p, yp))
                                l_middle.append((block2_E, yp))
                                l_middle.append((block2_E, y2p))
                    else: # NE b1W b2N
                        if block1_N < block2_N:
                            l_middle.append((x1p, block1_N))
                            l_middle.append((x2p, block1_N))
                        else:
                            if block1_W < block2_W:
                                l_middle.append((x1p, y2p))
                            else:
                                l_middle.append((block2_W, y1p))
                                l_middle.append((block2_W, y2p))
        elif direction2 == 'E':
            if delx_b12 > 0 or delx_b21 > 0 or dely_b12 > 0 or dely_b21 > 0:
                if west_of(c1p, c2p) == c2p: # NW and SW b1W b2E
                    delx = x1p - x2p
                    xp = x2p + int(delx/2)
                    l_middle.append((xp, y1p))
                    l_middle.append((xp, y2p))
                else:
                    if north_of(c1p, c2p) == c2p: # NE b1W b2E
                        if dely_b21 > 0:
                            yp = block2_S + int(dely_b21/2)
                            l_middle.append((x1p, yp))
                            l_middle.append((x2p, yp))
                        else:
                            yp = min(block1_N, block2_N)
                            if block2_W < x1p:
                                l_middle.append((block2_W, y1p))
                                l_middle.append((block2_W, yp))
                                l_middle.append((x2p, yp))
                            else:
                                l_middle.append((x1p, yp))
                                l_middle.append((x2p, yp))
                    else: # SE b1W b2E
                        if dely_b12 > 0:
                            yp = block1_S + int(dely_b12/2)
                            l_middle.append((x1p, yp))
                            l_middle.append((x2p, yp))
                        else:
                            yp = max(block1_S, block2_S)
                            if block2_W < x1p:
                                l_middle.append((block2_W, y1p))
                                l_middle.append((block2_W, yp))
                                l_middle.append((x2p, yp))
                            else:
                                l_middle.append((x1p, yp))
                                l_middle.append((x2p, yp))
            else:
                if x1p > x2p:
                    if y2p < y1p:
                        yp = min(block1_N, block2_N)
                        l_middle.append((x1p, yp))
                        l_middle.append((x2p, yp))
                    else:
                        yp = max(block1_S, block2_S)
                        l_middle.append((x1p, yp))
                        l_middle.append((x2p, yp))
        elif direction2 == 'W': # b1W b2W
            if dely_b12 > 0 or dely_b21 > 0 or delx_b12 > 0 or delx_b21 > 0:
                if block2_S < y1p or block2_N > y1p:
                    l_middle.append((x2p, y1p))
                else:
                    if block2_E > x1p:
                        l_middle.append((x2p, y1p))
                    else:
                        delx = x1p - block2_E
                        xp = block2_E + int(delx/2)
                        l_middle.append((xp, y1p))
                        l_middle.append((xp, block2_S))
                        l_middle.append((x2p, block2_S))
    elif direction1 == 'S':
        if direction2 == 'N':
            if dely_b12 > 0 or dely_b21 > 0 or delx_b12 > 0 or delx_b21 > 0:
                if south_of(c1p, c2p) == c2p: # SW and SE b1S b2N
                    dely = y2p - y1p
                    yp = y1p + int(dely/2)
                    l_middle.append((x1p, yp))
                    l_middle.append((x2p, yp))
                else:
                    if east_of(c1p, c2p) == c2p: # NE b1S b2N
                        if delx_b21 > 0:
                            xp = block1_E + int(delx_b21/2)
                            l_middle.append((xp, y1p))
                            l_middle.append((xp, y2p))
                        else:
                            xp = max(block1_E, block2_E)
                            if block2_S > y1p:
                                l_middle.append((x1p, block2_S))
                                l_middle.append((xp, block2_S))
                                l_middle.append((xp, y2p))
                            else:
                                l_middle.append((xp, y1p))
                                l_middle.append((xp, y2p))
                    else: # NW b1S b2N
                        if delx_b12 > 0:
                            xp = block2_E + int(delx_b12/2)
                            l_middle.append((xp, y1p))
                            l_middle.append((xp, y2p))
                        else:
                            xp = min(block1_W, block2_W)
                            if block2_S > y1p:
                                l_middle.append((x1p, block2_S))
                                l_middle.append((xp, block2_S))
                                l_middle.append((xp, y2p))
                            else:
                                l_middle.append((xp, y1p))
                                l_middle.append((xp, y2p))
            else:
                if y1p < y2p:
                    if x2p > x1p:
                        xp = min(block1_W, block2_W)
                        l_middle.append((xp, y1p))
                        l_middle.append((xp, y2p))
                    else:
                        xp = max(block1_E, block2_E)
                        l_middle.append((xp, y1p))
                        l_middle.append((xp, y2p))
        elif direction2 == 'S': # b1S b2S
            if dely_b12 > 0 or dely_b21 > 0 or delx_b12 > 0 or delx_b21 > 0:
                if block2_W > x1p or block2_E < x1p:
                    l_middle.append((x1p, y2p))
                else:
                    if block2_N < y1p:
                        l_middle.append((x1p, y2p))
                    else:
                        dely = block2_N - y1p
                        yp = block2_N - int(dely/2)
                        l_middle.append((x1p, yp))
                        l_middle.append((block2_W, yp))
                        l_middle.append((block2_W, y2p))
    elif direction1 == 'N':
        if direction2 == 'N': # b1N b2N
            if dely_b12 > 0 or dely_b21 > 0 or delx_b12 > 0 or delx_b21 > 0:
                if block2_W > x1p or block2_E < x1p:
                    l_middle.append((x1p, y2p))
                else:
                    if block2_S > y1p:
                        l_middle.append((x1p, y2p))
                    else:
                        dely = y1p - block2_S
                        yp = block2_S + int(dely/2)
                        l_middle.append((x1p, yp))
                        l_middle.append((block2_W, yp))
                        l_middle.append((block2_W, y2p))

    if not l_middle:
        l_corners = [(x1p, y2p), (x2p, y1p)]
        flag_added_corner = False
        for P in l_corners:
            xp, yp = P

            if direction1 in ('E', 'W'):
                p1 = (x1p - xp)*(x1p - xc1)
            else:
                p1 = (y1p - yp)*(y1p - yc1)

            if direction2 in ('E', 'W'):
                p2 = (x2p - xp)*(x2p - xc2)
            else:
                p2 = (y2p - yp)*(y2p - yc2)

            if p1 > 0 or p2 > 0:
                continue
            else:
                if p1 < 0 or p2 < 0:
                    l_middle.append(P)
                    flag_added_corner = True
                    break
                else:
                    continue

        if not flag_added_corner:

            if xc1 == xc2:
                c2p = [x2p, y1p]
            elif yc1 == yc2:
                c2p = [x1p, y2p]

    l_begin = [c1, c1p]
    l_end   = [c2p, c2]

    l_arrow1 = []
    l_arrow2 = []

    if arrow1:
        if direction1 == 'W':
            l_arrow1.append((xc1-arrow_h, yc1-arrow_w))
            l_arrow1.append((xc1        , yc1        ))
            l_arrow1.append((xc1-arrow_h, yc1+arrow_w))
        elif direction1 == 'E':
            l_arrow1.append((xc1+arrow_h, yc1-arrow_w))
            l_arrow1.append((xc1        , yc1        ))
            l_arrow1.append((xc1+arrow_h, yc1+arrow_w))
        elif direction1 == 'S':
            l_arrow1.append((xc1-arrow_w, yc1+arrow_h))
            l_arrow1.append((xc1        , yc1        ))
            l_arrow1.append((xc1+arrow_w, yc1+arrow_h))
        elif direction1 == 'N':
            l_arrow1.append((xc1-arrow_w, yc1-arrow_h))
            l_arrow1.append((xc1        , yc1        ))
            l_arrow1.append((xc1+arrow_w, yc1-arrow_h))

    if arrow2:
        if direction2 == 'W':
            l_arrow1.append((xc2-arrow_h, yc2-arrow_w))
            l_arrow1.append((xc2        , yc2        ))
            l_arrow1.append((xc2-arrow_h, yc2+arrow_w))
        elif direction2 == 'E':
            l_arrow1.append((xc2+arrow_h, yc2-arrow_w))
            l_arrow1.append((xc2        , yc2        ))
            l_arrow1.append((xc2+arrow_h, yc2+arrow_w))
        elif direction2 == 'S':
            l_arrow1.append((xc2-arrow_w, yc2+arrow_h))
            l_arrow1.append((xc2        , yc2        ))
            l_arrow1.append((xc2+arrow_w, yc2+arrow_h))
        elif direction2 == 'N':
            l_arrow1.append((xc2-arrow_w, yc2-arrow_h))
            l_arrow1.append((xc2        , yc2        ))
            l_arrow1.append((xc2+arrow_w, yc2-arrow_h))

    l = l_begin + l_middle + l_end

    if flag_flip:
        l.reverse()

    if l_arrow1:
        return l, l_arrow1
    elif l_arrow2:
        return l, l_arrow2
    else:
        return l, None



"""
Copyright 2008,2013 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from __future__ import absolute_import

from gi.repository import Gtk, Gdk, cairo

from .. import Constants

def get_color(color_code):
    color = Gdk.RGBA()
    color.parse(color_code)
    return color.red, color.green, color.blue, color.alpha
    # chars_per_color = 2 if len(color_code) > 4 else 1
    # offsets = range(1, 3 * chars_per_color + 1, chars_per_color)
    # return tuple(int(color_code[o:o + 2], 16) / 255.0 for o in offsets)

# fg colors

#HIGHLIGHT_COLOR = get_color('#00FFFF')
HIGHLIGHT_COLOR = get_color('#009DFF')
SELECT_COLOR = get_color('#A6D5FF')

#BORDER_COLOR = get_color('#616161')
BORDER_COLOR = get_color('#000000')
CONNECTOR_COLOR = get_color('#909090')

BORDER_COLOR_DISABLED = get_color('#888888')
FONT_COLOR = get_color('#000000')

E_PORT_COLOR = get_color('#D2F3FC')
F_PORT_COLOR = get_color('#fcb772')
B_PORT_COLOR = get_color('#1bcc6d')
PORT_NAME_COLOR = get_color('#0000ff')
GRID_LIGHT_COLOR = get_color('#c0c6cc')
GRID_DARK_COLOR = get_color('#a3a8ad')

# Missing blocks stuff
MISSING_BLOCK_BACKGROUND_COLOR = get_color('#FFF2F2')
MISSING_BLOCK_BORDER_COLOR = get_color('#FF0000')

# Flow graph color constants
FLOWGRAPH_BACKGROUND_COLOR = get_color('#FFFFFF')

COMMENT_BACKGROUND_COLOR = get_color('#F3F3F3')
FLOWGRAPH_EDGE_COLOR = COMMENT_BACKGROUND_COLOR

# Block color constants
BLOCK_ENABLED_COLOR = get_color('#F1ECFF')
#BLOCK_BUS_COLOR = get_color('#B0B0B0')
#BLOCK_BUS_COLOR = get_color('#0048ff')
BLOCK_BUS_COLOR = get_color('#3f75fc')
BLOCK_PAD_COLOR = get_color('#e0ffeb')

BLOCK_DUMMY_COLOR = get_color('#fdffcf')
#BLOCK_LABEL_COLOR = get_color('#0000AA')
BLOCK_LABEL_COLOR = get_color('#FF0000')
SHOW_TEXT_COLOR = get_color('#333333')
SHOW_PARAMETER_COLOR = get_color('#333333')

BLOCK_DISABLED_COLOR = get_color('#CCCCCC')
BLOCK_BYPASSED_COLOR = get_color('#F4FF81')

# Connection color constants
CONNECTION_ENABLED_COLOR = get_color('#000000')
CONNECTION_DISABLED_COLOR = get_color('#BBBBBB')
CONNECTION_ERROR_COLOR = get_color('#FF0000')

DEFAULT_DOMAIN_COLOR = get_color('#777777')




"""
Copyright 2008-2011,2015 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from __future__ import absolute_import

import numbers

from gi.repository import GLib
import cairo

from .canvas.colors import FLOWGRAPH_BACKGROUND_COLOR
from . import Constants

def scale_1(x, y, xmin, xmax, ymin, ymax):

    xmin0 = min(x)
    xmax0 = max(x)
    ymin0 = min(y)
    ymax0 = max(y)

    kx = (xmax - xmin)/(xmax0 - xmin0)
    ky = (ymax - ymin)/(ymax0 - ymin0)

    cx = xmax - kx*xmax0
    cy = ymax - ky*ymax0

    lx = []; ly = []
    for x0 in x:
        x1 = kx*x0 + cx
        lx.append(x1)
    for y0 in y:
        y1 = ky*y0 + cy
        ly.append(y1)

    return lx, ly

def get_rotated_coordinate(coor, rotation):
    """
    Rotate the coordinate by the given rotation.

    Args:
        coor: the coordinate x, y tuple
        rotation: the angle in degrees

    Returns:
        the rotated coordinates
    """
    # handles negative angles
    rotation = (rotation + 360) % 360
    if rotation not in Constants.POSSIBLE_ROTATIONS:
        raise ValueError('unusable rotation angle "%s"'%str(rotation))

    x, y = coor

    if rotation == 0:
        return x , y
    elif rotation == 90:
       return y, -x
    elif rotation == 180:
       return -x, -y
    elif rotation == 270:
       return -y, x

def get_rotated_coordinate_1(coor, delx, dely, rotation):

    rotation = (rotation + 360) % 360
    x, y = coor

    if rotation == 0:
        return x , y
    elif rotation == 90:
       return y, -x + delx
    elif rotation == 180:
       return -x + delx, -y + dely
    elif rotation == 270:
       return -y + dely, x

def get_rotated_coordinate_2(coor, delx, dely, rotation, mirror):

    x, y = coor
    if mirror == 'v':
        coor = (delx-x, y)
    elif mirror == 'h':
        coor = (x, dely-y)

    rotation = (rotation + 360) % 360
    x, y = coor

    if rotation == 0:
        return x , y
    elif rotation == 90:
       return y, -x + delx
    elif rotation == 180:
       return -x + delx, -y + dely
    elif rotation == 270:
       return -y + dely, x

def get_scaling_factors(t, rotation):

    k = {0:0, 90:1, 180:2, 270:3}[rotation]
    return t[k]

def get_rotated_angle_1(theta, rotation):
    alpha = int(theta - rotation)
    if alpha > 360: alpha -= 360
    if alpha < -360: alpha += 360
    return alpha

def get_rotated_angle_2(theta, rotation, mirror):

    if mirror == 'none':
        theta1 = theta
    elif mirror == 'v':
        if theta >= 0:
           theta1 = 180 - theta
        else:
           theta1 = -(180 + theta)
    elif mirror == 'h':
        theta1 = -theta

    alpha = int(theta1 - rotation)

    if alpha > 180: alpha -= 360
    if alpha < -180: alpha += 360

    return alpha

def get_angle_from_coordinates(p1, p2):
    """
    Given two points, calculate the vector direction from point1 to point2, directions are multiples of 90 degrees.

    Args:
        (x1,y1): the coordinate of point 1
        (x2,y2): the coordinate of point 2

    Returns:
        the direction in degrees
    """
    (x1, y1) = p1
    (x2, y2) = p2
    if y1 == y2:  # 0 or 180
        return 0 if x2 > x1 else 180
    else:  # 90 or 270
        return 270 if y2 > y1 else 90

def align_to_grid(coor, mode=round):

    def align(value):
        return int(mode(value / (1.0 * Constants.CANVAS_GRID_SIZE)) * Constants.CANVAS_GRID_SIZE)
    try:
        return [align(c) for c in coor]
    except TypeError:
        x = coor
        return align(coor)

def align_to_grid_1(coor, mode=round):

    def align(value):
        return int(mode(value / (2.0 * Constants.CANVAS_GRID_SIZE)) * 2*Constants.CANVAS_GRID_SIZE)
    try:
        return [align(c) for c in coor]
    except TypeError:
        x = coor
        return align(coor)

def num_to_str(num):
    """ Display logic for numbers """
    def eng_notation(value, fmt='g'):
        """Convert a number to a string in engineering notation.  E.g., 5e-9 -> 5n"""
        template = '{:' + fmt + '}{}'
        magnitude = abs(value)
        for exp, symbol in zip(range(9, -15-1, -3), 'GMk munpf'):
            factor = 10 ** exp
            if magnitude >= factor:
                return template.format(value / factor, symbol.strip())
        return template.format(value, '')

    if isinstance(num, numbers.Complex):
        num = complex(num)  # Cast to python complex
        if num == 0:
            return '0'
        output = eng_notation(num.real) if num.real else ''
        output += eng_notation(num.imag, '+g' if output else 'g') + 'j' if num.imag else ''
        return output
    else:
        return str(num)

def encode(value):
    """Make sure that we pass only valid utf-8 strings into markup_escape_text.

    Older versions of glib seg fault if the last byte starts a multi-byte
    character.
    """

    valid_utf8 = value

    return GLib.markup_escape_text(valid_utf8)

def make_screenshot(flow_graph, file_path, transparent_bg=False):
    if not file_path:
        return

    x_min, y_min, x_max, y_max = flow_graph.get_extents_1()

    padding = Constants.CANVAS_GRID_SIZE
    width = x_max - x_min + 2 * padding
    height = y_max - y_min + 2 * padding

    if file_path.endswith('.png'):
        psurf = cairo.ImageSurface(cairo.FORMAT_ARGB32, width, height)
    elif file_path.endswith('.pdf'):
        psurf = cairo.PDFSurface(file_path, width, height)
    elif file_path.endswith('.svg'):
        psurf = cairo.SVGSurface(file_path, width, height)
    else:
        raise ValueError('Unknown file format')

    cr = cairo.Context(psurf)

#   cr.set_line_join(cairo.LineJoin.ROUND)

    if not transparent_bg:
        cr.set_source_rgba(*FLOWGRAPH_BACKGROUND_COLOR)
        cr.rectangle(0, 0, width, height)
        cr.fill()

    cr.translate(padding - x_min, padding - y_min)

    flow_graph.create_labels(cr)
    flow_graph.create_shapes()

    flow_graph.draw_1(cr)

    if file_path.endswith('.png'):
        psurf.write_to_png(file_path)
    if file_path.endswith('.pdf') or file_path.endswith('.svg'):
        cr.show_page()
    psurf.finish()

def scale(coor, reverse=False):
    factor = Constants.DPI_SCALING if not reverse else 1 / Constants.DPI_SCALING
    return tuple(int(x * factor) for x in coor)

def scale_scalar(coor, reverse=False):
    factor = Constants.DPI_SCALING if not reverse else 1 / Constants.DPI_SCALING
    return int(coor * factor)



# Copyright 2008-2017 Free Software Foundation, Inc.
# This file is part of GNU Radio
#
# GNU Radio Companion is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# GNU Radio Companion is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from __future__ import absolute_import

import ast
import collections
import textwrap

from .. import Constants
from ..base import Element
from ..utils.descriptors import Evaluated, EvaluatedEnum, setup_names

from . import dtypes

attributed_str = type('attributed_str', (str,), {})

@setup_names
class Param(Element):

    is_param = True

    name = Evaluated(str, default='no name')
    dtype = EvaluatedEnum(Constants.PARAM_TYPE_NAMES, default='raw')

    hide = EvaluatedEnum('none all part')

    # region init
    def __init__(self, parent, id, label='', dtype='raw', default='',
                 options=None, option_labels=None, option_attributes=None,
                 category='', hide='none', **_):
        """Make a new param from nested data"""
        super(Param, self).__init__(parent)
        self.key = id
        self.name = label.strip() or id.title()
        self.category = category or Constants.DEFAULT_PARAM_TAB

        self.dtype = dtype
        self.value = self.default = str(default)

        self.options = self._init_options(options or [], option_labels or [],
                                          option_attributes or {})
        self.hide = hide or 'none'
        # end of args ########################################################

        self._evaluated = None
        self._stringify_flag = False
        self._lisitify_flag = False
        self.hostage_cells = set()
        self._init = False
        self.scale = {
            'E': 1e18,
            'P': 1e15,
            'T': 1e12,
            'G': 1e9,
            'M': 1e6,
            'k': 1e3,
            'm': 1e-3,
            'u': 1e-6,
            'n': 1e-9,
            'p': 1e-12,
            'f': 1e-15,
            'a': 1e-18,
        }
        self.scale_factor = None
        self.number = None

    def _init_options(self, values, labels, attributes):
        """parse option and option attributes"""
        options = collections.OrderedDict()
        options.attributes = collections.defaultdict(dict)

        padding = [''] * max(len(values), len(labels))
        attributes = {key: value + padding for key, value in iter(attributes.items())}

        for i, option in enumerate(values):
            # Test against repeated keys
            if option in options:
                raise KeyError('Value "{}" already exists in options'.format(option))
            # get label
            try:
                label = str(labels[i])
            except IndexError:
                label = str(option)
            # Store the option
            options[option] = label
            options.attributes[option] = {attrib: values[i] for attrib, values in iter(attributes.items())}

        default = next(iter(options)) if options else ''
        if not self.value:
            self.value = self.default = default

        if self.is_enum() and self.value not in options:
            self.value = self.default = default  # TODO: warn
            # raise ValueError('The value {!r} is not in the possible values of {}.'
            #                  ''.format(self.get_value(), ', '.join(self.options)))
        return options
    # endregion

    def __str__(self):
        return 'Param - {}({})'.format(self.name, self.key)

    def __repr__(self):
        return '{!r}.param[{}]'.format(self.parent, self.key)

    def is_enum(self):
        return self.get_raw('dtype') == 'enum'

    def get_value(self):
        value = self.value
        if self.is_enum() and value not in self.options:
            value = self.default
            self.set_value(value)
        return value

    def set_value(self, value):
        # Must be a string
        self.value = str(value)

    def set_default(self, value):
        if self.default == self.value:
            self.set_value(value)
        self.default = str(value)

    def rewrite(self):
        Element.rewrite(self)
        del self.name
        del self.dtype
        del self.hide

        self._evaluated = None
        try:
            self._evaluated = self.evaluate()
        except Exception as e:
            self.add_error_message(str(e))

        rewriter = getattr(dtypes, 'rewrite_' + self.dtype, None)
        if rewriter:
            rewriter(self)

    def validate(self):
        """
        Validate the param.
        The value must be evaluated and type must a possible type.
        """
        Element.validate(self)
        if self.dtype not in Constants.PARAM_TYPE_NAMES:
            self.add_error_message('Type "{}" is not a possible type.'.format(self.dtype))

        validator = dtypes.validators.get(self.dtype, None)
        if self._init and validator:
            try:
                validator(self)
            except dtypes.ValidateError as e:
                self.add_error_message(str(e))

    def get_evaluated(self):
        return self._evaluated

    def is_float(self, num):
        """
        Check if string can be converted to float.

        Returns:
            bool type
        """
        try:
            float(num)
            return True
        except ValueError:
            return False

    def evaluate(self):
        """
        Evaluate the value.

        Returns:
            evaluated type
        """
        self._init = True
        self._lisitify_flag = False
        self._stringify_flag = False
        dtype = self.dtype
        expr = self.get_value()
        scale_factor = self.scale_factor

        # ID and Enum types (not evaled)
        if dtype in ('id', 'name') or self.is_enum():
            if self.options.attributes:
                expr = attributed_str(expr)
                for key, value in self.options.attributes[expr].items():
                    setattr(expr, key, value)
            return expr

        # Numeric Types
        elif dtype in ('raw', 'complex', 'real', 'float', 'int', 'hex', 'bool'):
            if expr:
                try:
                    if isinstance(expr, str) and self.is_float(expr[:-1]):
                            scale_factor = expr[-1:]
                            if scale_factor in self.scale:
                                expr = str(float(expr[:-1])*self.scale[scale_factor])
                    value = self.parent_flowgraph.evaluate(expr)
                except Exception as e:
                    raise Exception('Value "{}" cannot be evaluated:\n{}'.format(expr, e))
            else:
                value = 0
            if dtype == 'hex':
                value = hex(value)
            elif dtype == 'bool':
                value = bool(value)
            return value

        # Numeric Vector Types
        # String Types
        elif dtype in ('string', 'file_open', 'file_save', '_multiline', '_multiline_python_external'):
            # Do not check if file/directory exists, that is a runtime issue
            try:
                # Do not evaluate multiline strings (code snippets or comments)
                if dtype not in ['_multiline','_multiline_python_external']:
                    value = self.parent_flowgraph.evaluate(expr)
                    if not isinstance(value, str):
                        raise Exception()
                else:
                    value = str(expr)
            except Exception:
                self._stringify_flag = True
                value = str(expr)
            if dtype == '_multiline_python_external':
                ast.parse(value)  # Raises SyntaxError
            return value
        # GUI Position/Hint
        # Import Type
        elif dtype == 'import':
            # New namespace
            n = dict()
            try:
                exec(expr, n)
            except ImportError:
                raise Exception('Import "{}" failed.'.format(expr))
            except Exception:
                raise Exception('Bad import syntax: "{}".'.format(expr))
            return [k for k in list(n.keys()) if str(k) != '__builtins__']

        else:
            raise TypeError('Type "{}" not handled'.format(dtype))

    def to_code(self):
        """
        Convert the value to code.
        For string and list types, check the init flag, call evaluate().
        This ensures that evaluate() was called to set the xxxify_flags.

        Returns:
            a string representing the code
        """
        self._init = True
        value = self.get_value()
        # String types
        if self.dtype in ('string', 'file_open', 'file_save', '_multiline', '_multiline_python_external'):
            if not self._init:
                self.evaluate()
            return repr(value) if self._stringify_flag else value

        else:
            return value

    def get_opt(self, item):
        return self.options.attributes[self.get_value()][item]

    # GUI Hint

    def get_all_params(self, dtype, key=None):
        """
        Get all the params from the flowgraph that have the given type and
        optionally a given key

        Args:
            dtype: the specified type
            key: the key to match against

        Returns:
            a list of params
        """
        params = []
        for block in self.parent_flowgraph.iter_enabled_blocks():
            params.extend(
                param for param in block.params.values()
                if param.dtype == dtype and (key is None or key == param.name)
            )
        return params



"""
Copyright (C) 2022 - Mahesh Patil <mbpatil@ee.iitb.ac.in>
This file is part of GSEIM.

GSEIM is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import sys
import os
import fcntl

def next_line_1(f, pos_in):
    string_list = []
    f.seek(pos_in)
    for i in range(500):
        pos = f.tell()
        line = f.readline()
        s2 = line.replace("\n","")
        s3 = s2.replace("="," ")
        s1 = s3.split()
        if (len(s1) == 0):
            continue
        if (s1[0] == '#'):
            continue
        if (len(string_list) == 0):
            for k in s1:
                string_list.append(k)
        else:
            if (s1[0] == '+'):
                for j in range(1,len(s1)):
                    string_list.append(s1[j])
            else:
                break
    return pos, string_list

def get_parms(filename, d_parms):

    f = open(os.path.expanduser(filename), 'r')
    pos_in = f.tell()
    pos_in, l_strings = next_line_1(f, pos_in)
    if l_strings[0] != 'begin_file':
        print('get_parms: expected begin_file. Halting...')
        sys.exit()
    while True:
        pos_in, l_strings = next_line_1(f, pos_in)
        if l_strings[0] == 'end_file':
            f.close()
            break
        if l_strings[0] != 'begin_parm':
            print('get_parms: expected begin_parm. Halting...')
            sys.exit()

        parm_category = 'none'
        if 'category' in l_strings:
            parm_category = l_strings[l_strings.index('category') + 1]

        pos_in, l_strings = next_line_1(f, pos_in)
        if l_strings[0] != 'keyword:':
            print('get_parms: expected keyword. Halting...')
            sys.exit()
        parm_name = l_strings[1]

        pos_in, l_strings = next_line_1(f, pos_in)
        if l_strings[0] != 'options:':
            print('get_parms: expected options. Halting...')
            sys.exit()
        l_options = l_strings[1:]

        pos_in, l_strings = next_line_1(f, pos_in)
        if l_strings[0] != 'default:':
            print('get_parms: expected default. Halting...')
            sys.exit()
        parm_default = l_strings[1]

        pos_in, l_strings = next_line_1(f, pos_in)
        if l_strings[0] != 'end_parm':
            print('get_parms: expected end_parm. Halting...')
            sys.exit()

        d_parms[parm_name] = {
            'options': l_options,
            'default': parm_default,
            'category': parm_category
        }    

def assign_parms_1(d_dialog, d_parms):
    for k, v in d_dialog.items():
        w = v['widget']

        if v['type_widget'] == 'entry':
            s = w.get_text().replace(' ', '')
            if s:
                d_parms[k] = s
            else:
                d_parms[k] = 'none'
        elif v['type_widget'] == 'checkbutton':
            d_parms[k] = 'yes' if w.get_active() else 'no'
        elif v['type_widget'] == 'combo':
            d_parms[k] = w.get_active_text()

def prepare_dict_1(d_parms, d_categories):

    for parm, d in d_parms.items():
        category = d['category']

        d_new = {'parm_name': parm, 'options': d['options'], 'default': d['default']}

        if category in d_categories.keys():
            d_categories[category].append(d_new)
        else:
            d_categories[category] = [d_new]

def find_nth(s_in, c, n):
#   https://stackoverflow.com/questions/30833409/python-deleting-the-first-2-lines-of-a-string

    start = s_in.find(c)
    while start >= 0 and n > 1:
        start = s_in.find(c, start+len(c))
        n -= 1
    return start

def is_running(pid):
    if os.path.isdir('/proc/{}'.format(pid)):
        return True
    return False

def get_sec(time_str):
#   https://stackoverflow.com/questions/6402812/how-to-convert-an-hmmss-time-string-to-seconds-in-python/6402934
    """Get Seconds from time."""
    h, m, s = time_str.split(':')
    return int(h)*3600 + int(m)*60 + int(s)

def non_block_read(output):
    ''' even in a thread, a normal read with block until the buffer is full '''
    fd = output.fileno()
    fl = fcntl.fcntl(fd, fcntl.F_GETFL)
    fcntl.fcntl(fd, fcntl.F_SETFL, fl | os.O_NONBLOCK)

    op = output.read()

    if op == None:
        return ''
    return op.decode('utf-8')

def python_code_to_list(filename, keyword1=None, keyword2=None):

    l1 = []

#   we will assume that there are no tab characters

    f = open(os.path.expanduser(filename), 'r')
    kw1_found = False

    for i in range(2000):
        x = f.readline()
        if not x: break

        if (keyword1 and kw1_found) or not keyword1:
            if keyword1 and keyword2 in x:
                break
            line = x.replace("\n","")
            if len(line.split()) == 0: continue
            line1 = line.split('#')[0]
            if len(line1.split()) == 0: continue
            l1.append(line1)
        else:
            if keyword1 and keyword1 in x:
                kw1_found = True
            continue

    f.close()

    if not kw1_found:
        print('python_code_to_list:', keyword1, 'not found. Halting...')
        sys.exit()

    if not l1:
        print('python_code_to_list: l1 is empty. Halting...')
        sys.exit()

    l2 = []
    flag_new = True

    for s in l1:
        if flag_new:
#           we will assume that \ appears at the end
#           (i.e., the user meant continuation)
            if '\\' in s:
                s_temp = s.replace('\\', ' ')
                flag_new = False
            else:
                l2.append(s.rstrip())
        else:
            if '\\' in s:
                s_temp += s.replace('\\', ' ')
            else:
                s_temp += s
                l2.append(s_temp.rstrip())
                flag_new = True

    n_indent = []

    for s in l2:
        if s.startswith(' '):
            n_sp = len(s) - len(s.lstrip())
            if n_sp % 4 != 0:
                print('python_code_to_list: indent must have 4, 8, .. spaces.')
                print('Check this statement:')
                print(s)
                sys.exit()
            n_indent.append(int(round(n_sp/4)))
        else:
            n_indent.append(0)

    if n_indent[0] != 0:
        print('python_code_to_list: indent must be 0 for the first line.')
        print('Check this statement:')
        print(l2[0])
        sys.exit()

    n1 = len(l2)
    if n1 == 1: return l2

    string_list = []
    flag_new = True
    s_temp = ''

    for i, s in enumerate(l2):
        if i == (n1-1):
            if flag_new:
                s_temp = s
            else:
                s_temp += '\n' + s
            string_list.append(s_temp.rstrip())
        else:
            if flag_new:
                s_temp = s
                if n_indent[i+1] > 0:
                    flag_new = False
                else:
                    string_list.append(s_temp.rstrip())
            else:
                s_temp += '\n' + s
                if n_indent[i+1] == 0:
                    flag_new = True
                    string_list.append(s_temp.rstrip())

    return string_list




# Copyright 2008-2015 Free Software Foundation, Inc.
# This file is part of GNU Radio
#
# GNU Radio Companion is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# GNU Radio Companion is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

from __future__ import absolute_import, print_function

import collections
import itertools
import sys
import types
import logging

from operator import methodcaller, attrgetter

from . import Messages, blocks
from .base import Element
from .utils import expr_utils

from ..SolveBlock import SolveBlock
from ..OutputBlock import OutputBlock

log = logging.getLogger(__name__)

class FlowGraph(Element):

    is_flow_graph = True

    def __init__(self, parent):
        """
        Make a flow graph from the arguments.

        Args:
            parent: a platforms with blocks and element factories

        Returns:
            the flow graph object
        """
        Element.__init__(self, parent)

        self.options_block = self.parent_platform.make_block(self, 'options')
        self.blocks = [self.options_block]

        self.connections = set()

        self.gparms = {}
        self.outvars = {}
        self.l_solve_blocks = []
        self.l_output_blocks = []
        self.d_slvparms = self.parent_platform.d_slvparms

        self._eval_cache = {}
        self.namespace = {}

        self.grc_file_path = ''

    def __str__(self):
        return 'FlowGraph - {}({})'.format(self.get_option('title'), self.get_option('id'))

    def get_variables(self):
        """
        Get a list of all variables (Python) in this flow graph namespace.
        Exclude parameterized variables.

        Returns:
            a sorted list of variable blocks in order of dependency (indep -> dep)
        """
        variables = [block for block in self.iter_enabled_blocks() if block.is_variable]
        return expr_utils.sort_objects(variables, attrgetter('name'), methodcaller('get_var_make'))

    def get_parameters(self):
        """
        Get a list of all parameterized variables in this flow graph namespace.

        Returns:
            a list of parameterized variables
        """
        parameters = [b for b in self.iter_enabled_blocks() if b.key == 'parameter']
        return parameters

    def iter_enabled_blocks(self):
        return (block for block in self.blocks)

    def get_enabled_blocks(self):
        """
        Get a list of all blocks that are enabled and not bypassed.

        Returns:
            a list of blocks
        """
        return list(self.iter_enabled_blocks())

    def get_enabled_connections(self):
        """
        Get a list of all connections that are enabled.

        Returns:
            a list of connections
        """
        return [connection for connection in self.connections]

    def get_option(self, key):
        """
        Get the option for a given key.
        The option comes from the special options block.

        Args:
            key: the param key for the options block

        Returns:
            the value held by that param
        """
        return self.options_block.params[key].get_evaluated()

    # Access Elements
    def get_block(self, name):
        for block in self.blocks:
            if block.name == name:
                return block
        raise KeyError('No block with name {!r}'.format(name))

    def get_elements(self):
        elements = list(self.blocks)
        elements.extend(self.connections)
        return elements

    def children(self):
        return itertools.chain(self.iter_enabled_blocks(), self.connections)

    def rewrite(self):
        """
        Flag the namespace to be renewed.
        """
        self.renew_namespace()
        Element.rewrite(self)

    def renew_namespace(self):
        namespace = {}
        # Load imports

        # Load parameters
        np = {}  # params don't know each other
        for parameter_block in self.get_parameters():
            try:
                value = eval(parameter_block.params['value'].to_code(), namespace)
                np[parameter_block.name] = value
            except Exception:
                log.exception('Failed to evaluate parameter block {0}'.format(parameter_block.name), exc_info=True)
                pass
        namespace.update(np)  # Merge param namespace

        self.namespace.clear()
        self._eval_cache.clear()
        self.namespace.update(namespace)

    def evaluate(self, expr, namespace=None, local_namespace=None):
        """
        Evaluate the expression.
        """
        # Evaluate
        if not expr:
            raise Exception('Cannot evaluate empty statement.')
        if namespace is not None:
            return eval(expr, namespace, local_namespace)
        else:
            return self._eval_cache.setdefault(expr, eval(expr, self.namespace, local_namespace))

    # Add/remove stuff

    def new_block(self, block_id, **kwargs):
        """
        Get a new block of the specified key.
        Add the block to the list of elements.

        Args:
            block_id: the block key

        Returns:
            the new block or None if not found
        """
        if block_id == 'options':
            return self.options_block
        try:
            block = self.parent_platform.make_block(self, block_id, **kwargs)
            self.blocks.append(block)
        except KeyError:
            block = None

        return block

    def connect(self, porta, portb):
        """
        Create a connection between porta and portb.

        Args:
            porta: a port
            portb: another port
        @throw Exception bad connection

        Returns:
            the new connection
        """

        connection = self.parent_platform.Connection(
            parent=self, source=porta, sink=portb)
        self.connections.add(connection)
            
        return connection

    def disconnect(self, *ports):
        to_be_removed = [con for con in self.connections
                         if any(port in con for port in ports)]
        for con in to_be_removed:
            self.remove_element(con)

    def remove_element(self, element):
        """
        Remove the element from the list of elements.
        If the element is a port, remove the whole block.
        If the element is a block, remove its connections.
        If the element is a connection, just remove the connection.
        """
        if element is self.options_block:
            return

        if element.is_port:
            element = element.parent_block  # remove parent block

        if element in self.blocks:
            # Remove block, remove all involved connections
            self.disconnect(*element.ports())
            self.blocks.remove(element)

        elif element in self.connections:
            self.connections.remove(element)

    # Import/Export Methods
    def export_data(self):
        """
        Export this flow graph to nested data.
        Export all block and connection data.

        Returns:
            a nested data odict
        """
        def block_order(b):
            return not b.is_variable, b.name  # todo: vars still first ?!?

        data = collections.OrderedDict()
        data['options'] = self.options_block.export_data()
        data['blocks'] = [b.export_data() for b in sorted(self.blocks, key=block_order)
                          if b is not self.options_block]
        data['connections'] = sorted(c.export_data() for c in self.connections)

        data['gparms'] = self.gparms

#       Assume the following form for data['outvars']
#       - for connection: { name: ['wire', [xx, xx, xx, xx]], ...}
#         (where the list inside describes the connection
#       - for block parm: { name: ['outvar', [xx, yy]], ...}
#         (xx: block/subckt name, yy: which outvar of that block/subckt)

        data['outvars'] = self.outvars

        l_data_solve = []
        l_data_output = []

        for slv in self.l_solve_blocks:
            l_data_solve.append(slv.block_to_dict())

        for out in self.l_output_blocks:
            l_data_output.append(out.block_to_dict())

        data['solve_blocks'] = l_data_solve
        data['output_blocks'] = l_data_output

        return data

    def import_data(self, data):
        """
        Import blocks and connections into this flow graph.
        Clear this flow graph of all previous blocks and connections.
        Any blocks or connections in error will be ignored.

        Args:
            data: the nested data odict
        """
        # Remove previous elements
        del self.blocks[:]
        self.connections.clear()

        if 'gparms' in data.keys():
            self.gparms = data['gparms']
        else:
            self.gparms = {}

        if 'outvars' in data.keys():
            self.outvars = data['outvars']
        else:
            self.outvars = {}

        self.l_solve_blocks = []
        self.l_output_blocks = []

        if 'solve_blocks' in data.keys():
            if len(data['solve_blocks']) != 0:
                for d in data['solve_blocks']:

                    slv = SolveBlock(self.d_slvparms, '', '', d)
                    self.l_solve_blocks.append(slv)

        if 'output_blocks' in data.keys():
            if len(data['output_blocks']) != 0:
                for d in data['output_blocks']:
                    out = OutputBlock({}, '', '', d)
                    self.l_output_blocks.append(out)

        # build the blocks
        self.options_block.import_data(name='', **data.get('options', {}))
        self.blocks.append(self.options_block)

        for block_data in data.get('blocks', []):
            block_id = block_data['id']
            block = (
                self.new_block(block_id) or
                self.new_block(block_id='_dummy', missing_block_id=block_id, **block_data)
            )

            block.import_data(**block_data)

        self.rewrite() 

        # build the connections
        def verify_and_get_port(key, block, dir):

            ports = {
              'sink'    : block.sinks,
              'source'  : block.sources,
              'e_left'  : block.e_lefts,
              'e_right' : block.e_rights,
              'e_top'   : block.e_tops,
              'e_bottom': block.e_bottoms,
              'b_left'  : block.b_lefts,
              'b_right' : block.b_rights,
              'b_top'   : block.b_tops,
              'b_bottom': block.b_bottoms,
            }[dir]

            for port in ports:

                if key == port.key:
                    break

            else:
                if block.is_dummy_block:
                    port = block.add_missing_port(key, dir)
                else:
                    raise LookupError('%s key %r not in %s block keys' % (dir, key, dir))

            return port

        had_connect_errors = False
        _blocks = {block.name: block for block in self.blocks}

        try:
            # TODO: Add better error handling if no connections exist in the flowgraph file.
            for src_blk_id, src_port_id, snk_blk_id, snk_port_id in data.get('connections', []):

                source_block = _blocks[src_blk_id]
                sink_block = _blocks[snk_blk_id]

                # fix old, numeric message ports keys

                # build the connection

                if src_port_id.startswith('e'):

                    src_port_id_1 = src_port_id[2:].strip()
                    snk_port_id_1 = snk_port_id[2:].strip()

                    src_prefix = src_port_id[0:2]
                    snk_prefix = snk_port_id[0:2]

                    src_port_type = {
                       'el' : 'e_left',
                       'er' : 'e_right',
                       'et' : 'e_top',
                       'eb' : 'e_bottom',
                    }[src_prefix]

                    snk_port_type = {
                       'el' : 'e_left',
                       'er' : 'e_right',
                       'et' : 'e_top',
                       'eb' : 'e_bottom',
                    }[snk_prefix]

                    source_port = verify_and_get_port(src_port_id_1, source_block, src_port_type)
                    sink_port = verify_and_get_port(snk_port_id_1, sink_block, snk_port_type)
                elif src_port_id.startswith('b'):

                    src_port_id_1 = src_port_id[2:].strip()
                    snk_port_id_1 = snk_port_id[2:].strip()

                    src_prefix = src_port_id[0:2]
                    snk_prefix = snk_port_id[0:2]

                    src_port_type = {
                       'bl' : 'b_left',
                       'br' : 'b_right',
                       'bt' : 'b_top',
                       'bb' : 'b_bottom',
                    }[src_prefix]

                    snk_port_type = {
                       'bl' : 'b_left',
                       'br' : 'b_right',
                       'bt' : 'b_top',
                       'bb' : 'b_bottom',
                    }[snk_prefix]

                    source_port = verify_and_get_port(src_port_id_1, source_block, src_port_type)
                    sink_port = verify_and_get_port(snk_port_id_1, sink_block, snk_port_type)
                else:
                    source_port = verify_and_get_port(src_port_id, source_block, 'source')
                    sink_port = verify_and_get_port(snk_port_id, sink_block, 'sink')

                self.connect(source_port, sink_port)

        except (KeyError, LookupError) as e:
            Messages.send_error_load(
                'Connection between {}({}) and {}({}) could not be made.\n\t{}'.format(
                    src_blk_id, src_port_id, snk_blk_id, snk_port_id, e))
            had_connect_errors = True

        for block in self.blocks:
            if block.is_dummy_block :
                block.rewrite()      # Make ports visible
                # Flowgraph errors depending on disabled blocks are not displayed
                # in the error dialog box
                # So put a message into the Property window of the dummy block
                block.add_error_message('Block id "{}" not found.'.format(block.key))

        self.rewrite()  # global rewrite
        return had_connect_errors

# edited by mbp



"""
Copyright (C) 2022 - Mahesh Patil <mbpatil@ee.iitb.ac.in>
This file is part of GSEIM.

GSEIM is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

class AddShowParm(Gtk.Dialog):
    def __init__(self, parent, l_parm_names):
        Gtk.Dialog.__init__(self, title="Show Parameter", transient_for=parent, flags=0)
        self.add_buttons(
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OK, Gtk.ResponseType.OK
        )
        self.set_default_size(400, 500)
        self.set_position(Gtk.WindowPosition.CENTER)

        box = self.get_content_area()
        box.set_spacing(30)
        box.set_margin_top(5)

        grid = Gtk.Grid()
        grid.set_row_spacing(10)
        grid.set_column_spacing(10)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolled_window.show()
        self.vbox.pack_start(scrolled_window, True, True, 0)
        scrolled_window.add(grid)

        self.l_buttons = []

        for i, parm_name in enumerate(l_parm_names):
            if i == 0:
                button1 = Gtk.RadioButton(label=parm_name)
                button1.set_active(False)
                self.l_buttons.append(button1)
            else:
                button1 = Gtk.RadioButton.new_from_widget(self.l_buttons[i-1])
                button1.set_label(parm_name)
                self.l_buttons.append(button1)

        label_1 = Gtk.Label(label = ' Select parameter ', xalign=0.5)

        grid.add(label_1)

        grid.attach_next_to(self.l_buttons[0],
           label_1, Gtk.PositionType.BOTTOM, 1, 1)

        for i in range(1, len(l_parm_names)):
            grid.attach_next_to(self.l_buttons[i], self.l_buttons[i-1],
               Gtk.PositionType.BOTTOM, 1, 1)

        box.add(grid)
        self.show_all()



from .utils import Spec, expand, str_

OPTIONS_SCHEME = expand(
    parameters=Spec(types=dict, required=False, item_scheme=(str_, str_)),
    states=Spec(types=dict, required=False, item_scheme=(str_, str_)),
)

BLOCK_SCHEME = expand(
    name=str_,
    id=str_,
    outvars=str_,
    **OPTIONS_SCHEME
)

FLOW_GRAPH_SCHEME = expand(
    options=Spec(types=dict, required=False, item_scheme=OPTIONS_SCHEME),
    blocks=Spec(types=dict, required=False, item_scheme=BLOCK_SCHEME),
    connections=list,

)



"""
Copyright (C) 2022 - Mahesh Patil <mbpatil@ee.iitb.ac.in>
This file is part of GSEIM.

GSEIM is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import gi

gi.require_version("Gtk", "3.0")
from gi.repository import Gtk

class AddOutvarConnection(Gtk.Dialog):
    def __init__(self, parent, ovname_next, ov_value):
        Gtk.Dialog.__init__(self, title="Add Outvar", transient_for=parent, flags=0)
        self.add_buttons(
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OK, Gtk.ResponseType.OK
        )
        self.set_default_size(400, 100)
        self.set_position(Gtk.WindowPosition.CENTER)

        box = self.get_content_area()
        box.set_spacing(30)
        box.set_margin_top(5)

        grid = Gtk.Grid()
        grid.set_row_spacing(10)
        grid.set_column_spacing(10)

        label_name = Gtk.Label(label = "  Name ", xalign=1)
        label_value = Gtk.Label(label = "  Value ", xalign=1)

        self.entry_name = Gtk.Entry(text=ovname_next)
        self.entry_value = Gtk.Entry(text=ov_value)
        self.entry_value.set_width_chars(50)

        self.entry_value.set_property("editable", False)

        grid.add(label_name)
        grid.attach(self.entry_name, 1, 0, 1, 1)
        grid.attach_next_to(label_value, label_name, Gtk.PositionType.BOTTOM, 1, 1)
        grid.attach_next_to(self.entry_value, label_value, Gtk.PositionType.RIGHT, 1, 1)

        box.add(grid)
        self.show_all()

class AddOutvarOutvar(Gtk.Dialog):
    def __init__(self, parent, ovname_next, l_ov_values):
        Gtk.Dialog.__init__(self, title="Add Outvar", transient_for=parent, flags=0)
        self.add_buttons(
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OK, Gtk.ResponseType.OK
        )
        self.set_default_size(400, 100)
        self.set_position(Gtk.WindowPosition.CENTER)

        box = self.get_content_area()
        box.set_spacing(30)
        box.set_margin_top(5)

        grid = Gtk.Grid()
        grid.set_row_spacing(10)
        grid.set_column_spacing(10)

        self.l_buttons = []

        for i,ov_value in enumerate(l_ov_values):
            if i == 0:
                button1 = Gtk.RadioButton(label=ov_value)
                button1.set_active(False)
                self.l_buttons.append(button1)
            else:
                button1 = Gtk.RadioButton.new_from_widget(self.l_buttons[i-1])
                button1.set_label(ov_value)
                self.l_buttons.append(button1)

        label_name = Gtk.Label(label = "  Name ", xalign=1)
        self.entry_name = Gtk.Entry(text=ovname_next)

        grid.add(label_name)
        grid.attach(self.entry_name, 1, 0, 1, 1)

        grid.attach_next_to(self.l_buttons[0],
           self.entry_name, Gtk.PositionType.BOTTOM, 1, 1)

        for i in range(1, len(l_ov_values)):
            grid.attach_next_to(self.l_buttons[i], self.l_buttons[i-1],
               Gtk.PositionType.BOTTOM, 1, 1)

        box.add(grid)
        self.show_all()

class DelOutvar(Gtk.Dialog):
    def __init__(self, parent, ov_dict):

        Gtk.Dialog.__init__(self, title="Delete Outvars", transient_for=parent, flags=0)
        self.add_buttons(
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OK, Gtk.ResponseType.OK
        )
        self.set_default_size(450, 200)
        self.set_position(Gtk.WindowPosition.CENTER)

        grid = Gtk.Grid()
        grid.set_row_spacing(10)
        grid.set_column_spacing(10)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolled_window.show()
        self.vbox.pack_start(scrolled_window, True, True, 0)
        scrolled_window.add(grid)

        l_names = list(ov_dict.keys())
        l_values = list(ov_dict.values())
        self.name = []
        self.tick = []

        for i in range(len(l_names)):
            s1 = l_names[i] + ': ' + str(l_values[i])
            self.name.append(Gtk.Label(label=s1, xalign=0.0))
            self.tick.append(Gtk.CheckButton())

        grid.add(self.name[0])
        grid.attach(self.tick[0], 1, 0, 1, 1)

        for i in range(1, len(l_names)):
            grid.attach_next_to(self.name[i], self.name[i-1], Gtk.PositionType.BOTTOM, 1, 1)
            grid.attach_next_to(self.tick[i], self.name[i], Gtk.PositionType.RIGHT, 1, 1)

        self.show_all()

class EditOutvar(Gtk.Dialog):
    def __init__(self, parent, ov_dict):

        l_names = []

        for v in ov_dict.values():
            l_names.append(v[1])

        l_values = list(ov_dict.keys())

        Gtk.Dialog.__init__(self, title="Edit Outvar Names", transient_for=parent, flags=0)
        self.add_buttons(
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OK, Gtk.ResponseType.OK
        )
        self.set_default_size(550, 200)
        self.set_position(Gtk.WindowPosition.CENTER)

        grid = Gtk.Grid()
        grid.set_row_spacing(10)
        grid.set_column_spacing(10)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolled_window.show()
        self.vbox.pack_start(scrolled_window, True, True, 0)
        scrolled_window.add(grid)

        self.name = []
        self.value = []

        for i in range(len(l_names)):
            self.name.append(Gtk.Label(label=l_names[i], xalign=0))
            self.value.append(Gtk.Entry(text=l_values[i]))

        grid.add(self.name[0])
        grid.attach(self.value[0], 1, 0, 1, 1)

        for i in range(1, len(l_names)):
            grid.attach_next_to(self.name[i], self.name[i-1], Gtk.PositionType.BOTTOM, 1, 1)
            grid.attach_next_to(self.value[i], self.name[i], Gtk.PositionType.RIGHT, 1, 1)

        self.show_all()




"""
Copyright (C) 2022 - Mahesh Patil <mbpatil@ee.iitb.ac.in>
This file is part of GSEIM.

GSEIM is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <https://www.gnu.org/licenses/>.
"""

import gi
import sys
gi.require_version('Gtk', '3.0')
from gi.repository import Gtk
from ..core.utils import gutils as gu

class AddSolveBlock(Gtk.Dialog):
    def __init__(self, parent):
        Gtk.Dialog.__init__(self, title='Add Solve Block', transient_for=parent, flags=0)
        self.add_buttons(
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OK, Gtk.ResponseType.OK
        )
        self.set_default_size(150, 100)
        self.set_position(Gtk.WindowPosition.CENTER)

        box = self.get_content_area()

        box.set_spacing(30)
        box.set_margin_top(5)

        grid = Gtk.Grid()

        grid.set_row_spacing(10)
        grid.set_column_spacing(10)

        label_name = Gtk.Label(label = '  Name ', xalign=1)
        label_index = Gtk.Label(label = '  Index ', xalign=1)

        self.entry_name = Gtk.Entry(text='')
        self.entry_value = Gtk.Entry(text='0')

        grid.add(label_name)
        grid.attach(self.entry_name, 1, 0, 1, 1)
        grid.attach_next_to(label_index, label_name, Gtk.PositionType.BOTTOM, 1, 1)
        grid.attach_next_to(self.entry_value, label_index, Gtk.PositionType.RIGHT, 1, 1)

        box.add(grid)
        self.show_all()

class DelSolveBlock(Gtk.Dialog):
    def __init__(self, parent, l_solve_blocks):
#       l_solve_blocks is a list of solve blocks, each solve block is a dict
#       Note: do not need to add warning if l_solve_blocks is empty; that would be
#         done in the calling function

        Gtk.Dialog.__init__(self, title='Delete Solve Blocks', transient_for=parent, flags=0)
        self.add_buttons(
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OK, Gtk.ResponseType.OK
        )
        self.set_default_size(150, 100)
        self.set_position(Gtk.WindowPosition.CENTER)

        box = self.get_content_area()
        box.set_spacing(30)
        box.set_margin_top(5)

        grid = Gtk.Grid()
        grid.set_row_spacing(10)
        grid.set_column_spacing(10)
        grid.props.halign = Gtk.Align.CENTER

        l_names = []

#       if the solve block name is slv_1 and index is 0, we will display
#       'slv_1 (0)' as the label
#       We will treat index as a string, not int.

        for slv in l_solve_blocks:
            l_names.append(slv.name + ' (' + slv.index + ')')

        self.name = []
        self.tick = []

        for i in range(len(l_names)):
            self.name.append(Gtk.Label(label=l_names[i], xalign=0.5))
            self.tick.append(Gtk.CheckButton())

        grid.add(self.name[0])
        grid.attach(self.tick[0], 1, 0, 1, 1)

        for i in range(1, len(l_names)):
            grid.attach_next_to(self.name[i], self.name[i-1], Gtk.PositionType.BOTTOM, 1, 1)
            grid.attach_next_to(self.tick[i], self.name[i], Gtk.PositionType.RIGHT, 1, 1)

        box.add(grid)
        self.show_all()

class PickSolveBlock(Gtk.Dialog):
    def __init__(self, parent, l_solve_blocks):
#       Note: do not need to add warning if l_solve_blocks is empty; that would be
#         done in the calling function

        Gtk.Dialog.__init__(self, title='Pick Solve Block for Editing',
            transient_for=parent, flags=0)
        self.add_buttons(
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OK, Gtk.ResponseType.OK
        )
        self.set_default_size(150, 100)
        self.set_position(Gtk.WindowPosition.CENTER)

        box = self.get_content_area()
        box.set_spacing(30)
        box.set_margin_top(5)

        grid = Gtk.Grid()
        grid.set_row_spacing(10)
        grid.set_column_spacing(10)
        grid.props.halign = Gtk.Align.CENTER

        l_names = []
        self.l_buttons = []

        for slv in l_solve_blocks:
            l_names.append(slv.name + ' (' + slv.index + ')')

        self.l_buttons.append(Gtk.RadioButton(label=l_names[0]))

        for i in range(1, len(l_names)):
            button1 = Gtk.RadioButton.new_from_widget(self.l_buttons[i-1])
            button1.set_label(l_names[i])
            self.l_buttons.append(button1)

        grid.add(self.l_buttons[0])

        for i in range(1, len(l_names)):
            grid.attach_next_to(self.l_buttons[i], self.l_buttons[i-1],
               Gtk.PositionType.BOTTOM, 1, 1)

        box.add(grid)
        self.show_all()

class EditSolveBlock(Gtk.Dialog):
#   Present the selected solve block for editing:

    def __init__(self, parent, blk, d_slv_categories):
#       blk is the selected solve block
        Gtk.Dialog.__init__(self,
            title=blk.name,
            transient_for=parent, flags=0)
        self.add_buttons(
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OK, Gtk.ResponseType.OK
        )
        self.set_default_size(400, 800)
        self.set_position(Gtk.WindowPosition.CENTER)

        grid = Gtk.Grid()
        grid.set_row_spacing(0)
        grid.set_column_spacing(0)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolled_window.show()
        self.vbox.pack_start(scrolled_window, True, True, 0)
        scrolled_window.add(grid)

        d_widgets = {}
        set1 = {'yes', 'no'}

        self.d_widgets_1 = {}

        for d in d_slv_categories['none']:
            parm_name = d['parm_name']
            l_options = d['options']

            if parm_name in blk.d_parms.keys():
                s_value = blk.d_parms[parm_name]
            else:
                s_value = d['default']

            l_widgets = []

#           if key is initial_sol_file, then we need special treatment
#           button/entry with button connected to a function which will
#           use Gtk.FileChooserDialog

            if parm_name == 'initial_sol_file':
                button1 = Gtk.Button(label="init sol file", xalign=0)
                button1.connect("clicked", self.on_file_clicked)
                l_widgets.append([button1, [0, 0, 1, 1]])
            else:
                label_1 = Gtk.Label(label = '  ' + parm_name + '  ', xalign=0)
                l_widgets.append([label_1, [0, 0, 1, 1]])

            if l_options[0] == 'none':
                if s_value == 'none':
                    entry_1 = Gtk.Entry(text='')
                else:
                    entry_1 = Gtk.Entry(text=s_value)

                self.d_widgets_1[parm_name] = {'type_widget': 'entry', 'widget':entry_1}
                l_widgets.append([entry_1, [1, 0, 1, 1]])
            else:
                if set(l_options) == set1:
                    button_1 = Gtk.CheckButton()
                    button_1.set_active(s_value == 'yes')
                    button_1.set_label('yes/no')
                    l_widgets.append([button_1, [1, 0, 1, 1]])
                    self.d_widgets_1[parm_name] = {'type_widget': 'checkbutton', 'widget':button_1}
                else:
                    combo_1 = Gtk.ComboBoxText()
                    combo_1.set_entry_text_column(0)
                    for x in l_options:
                        combo_1.append_text(x)
                    combo_1.set_active(l_options.index(s_value))
                    self.d_widgets_1[parm_name] = {'type_widget': 'combo', 'widget':combo_1}

                    l_widgets.append([combo_1, [1, 0, 1, 1]])

            d_widgets[parm_name] = l_widgets

        l_keys = []
        for d in d_slv_categories['none']:
            parm_name = d['parm_name']
            l_keys.append(parm_name)

        for k,l in d_slv_categories.items():
            if k != 'none':
                button1 = Gtk.Button(label=k)
                button1.connect('clicked', self.solve_options_1, k, l, blk)
                self.d_widgets_1[k] = {'type_widget': 'button', 'widget':button1}
                l_widgets = [[button1, [0, 0, 2, 1]]]
                d_widgets[k] = l_widgets
                l_keys.append(k)

        disp = 0

        for k in l_keys:
            for l0 in d_widgets[k]:
                l1 = l0[1]
                grid.attach(l0[0], l1[0], l1[1] + disp, l1[2], l1[3])
            d0 = max(map(lambda x: x[1][1], d_widgets[k])) + 1
                    
            disp += d0

        self.show_all()

    def on_file_clicked(self, widget):
        _dialog = Gtk.FileChooserDialog(
            title="Please choose a file", parent=self,
            action=Gtk.FileChooserAction.OPEN
        )
        _dialog.add_buttons(
            Gtk.STOCK_CANCEL,
            Gtk.ResponseType.CANCEL,
            Gtk.STOCK_OPEN,
            Gtk.ResponseType.OK,
        )

        _filter_gsol = Gtk.FileFilter()
        _filter_gsol.set_name("gsol files")
        _filter_gsol.add_pattern("*.gsol")
        _dialog.add_filter(_filter_gsol)

        response = _dialog.run()
        if response == Gtk.ResponseType.OK:
            print("Open clicked")
            filename = _dialog.get_filename()
            print("File selected: " + _dialog.get_filename())
            self.d_widgets_1['initial_sol_file']['widget'].set_text(filename)
        elif response == Gtk.ResponseType.CANCEL:
            print("Cancel clicked")

        _dialog.destroy()

    def solve_options_1(self, widget, category, l_dict, blk):

        dialog2 = EditSolveGroup(self, category, l_dict, blk)

        response = dialog2.run()
        if response == Gtk.ResponseType.OK:
            print('solve_options_1: dialog2: The OK button was clicked')
            gu.assign_parms_1(dialog2.d_widgets_1, blk.d_parms)
        elif response == Gtk.ResponseType.CANCEL:
            print('solve_options_1: dialog2: The Cancel button was clicked')

        dialog2.destroy()
        return

class EditSolveGroup(Gtk.Dialog):
#   Present the statements from a solve block group (such as x_nr)

    def __init__(self, parent, category, l_dict, blk):
        Gtk.Dialog.__init__(self,
            title=blk.name + ': ' + category,
            transient_for=parent, flags=0)
        self.add_buttons(
            Gtk.STOCK_CANCEL, Gtk.ResponseType.CANCEL, Gtk.STOCK_OK, Gtk.ResponseType.OK
        )
        self.set_default_size(400, 800)
        self.set_position(Gtk.WindowPosition.CENTER)

        grid = Gtk.Grid()
        grid.set_row_spacing(0)
        grid.set_column_spacing(0)

        scrolled_window = Gtk.ScrolledWindow()
        scrolled_window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
        scrolled_window.show()
        self.vbox.pack_start(scrolled_window, True, True, 0)

        scrolled_window.add(grid)

        d_widgets = {}
        set1 = {'yes', 'no'}

        self.d_widgets_1 = {}

        for d in l_dict:
            parm_name = d['parm_name']
            l_options = d['options']
            if parm_name not in blk.d_parms.keys():
                print('EditSolveGroup:', parm_name, 'not found in blk.d_parms.keys(). Halting...')
                sys.exit()

            s_value = blk.d_parms[parm_name]

            l_widgets = []

            label_1 = Gtk.Label(label = '  ' + parm_name + '  ', xalign=0)
            l_widgets.append([label_1, [0, 0, 1, 1]])

            if l_options[0] == 'none':
                if s_value == 'none':
                    entry_1 = Gtk.Entry(text='')
                else:
                    entry_1 = Gtk.Entry(text=s_value)

                self.d_widgets_1[parm_name] = {'type_widget': 'entry', 'widget':entry_1}
                l_widgets.append([entry_1, [1, 0, 1, 1]])
            else:
                if set(l_options) == set1:
                    button_1 = Gtk.CheckButton()
                    button_1.set_active(s_value == 'yes')
                    button_1.set_label('yes/no')
                    l_widgets.append([button_1, [1, 0, 1, 1]])
                    self.d_widgets_1[parm_name] = {'type_widget': 'checkbutton', 'widget':button_1}
                else:
                    combo_1 = Gtk.ComboBoxText()
                    combo_1.set_entry_text_column(0)
                    for x in l_options:
                        combo_1.append_text(x)
                    combo_1.set_active(l_options.index(s_value))
                    self.d_widgets_1[parm_name] = {'type_widget': 'combo', 'widget':combo_1}

                    l_widgets.append([combo_1, [1, 0, 1, 1]])

            d_widgets[parm_name] = l_widgets

        l_keys = []

        for d in l_dict:
            parm_name = d['parm_name']
            l_keys.append(parm_name)

        disp = 0

        for k in l_keys:
            for l0 in d_widgets[k]:
                l1 = l0[1]
                grid.attach(l0[0], l1[0], l1[1] + disp, l1[2], l1[3])
            d0 = max(map(lambda x: x[1][1], d_widgets[k])) + 1
                    
            disp += d0

        self.show_all()



"""
Copyright 2007-2011 Free Software Foundation, Inc.
This file is part of GNU Radio

GNU Radio Companion is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

GNU Radio Companion is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA
"""

from __future__ import absolute_import, print_function

import logging
import os
import time
import subprocess

from gi.repository import Gtk, Gio, GLib, GObject
from getpass import getuser

from gi.repository import Pango

from . import Constants, Dialogs, Actions, FileDialogs, Utils, Bars

from .MainWindow import MainWindow
from .PropsDialog import PropsDialog

from ..core import Messages

import psutil
import sys
import shutil
from . import gparms as gp
from . import outvars as ov
from . import showparm as sparm
from . import solveblocks as slvblk
from . import outputblocks as outblk
from . import displayname as disp
from . import runprocess as runp
from . import showlist as slist
from .canvas.connection import Connection
from ..SolveBlock import SolveBlock
from ..OutputBlock import OutputBlock
from ..core.utils import gutils as gu
from .BlockTreeWindow import show_doc_1

log = logging.getLogger(__name__)

class Application(Gtk.Application):
    """
    The action handler will setup all the major window components,
    and handle button presses and flow graph operations from the GUI.
    """

    def __init__(self, file_paths, platform):
        Gtk.Application.__init__(self)
        """
        Application constructor.
        Create the main window, setup the message handler, import the preferences,
        and connect all of the action handlers. Finally, enter the gtk main loop and block.

        Args:
            file_paths: a list of flow graph file passed from command line
            platform: platform module
        """
        self.dict_outvars = platform.dict_outvars.copy()
        self.d_slvparms = platform.d_slvparms.copy()
        self.d_slv_categories = platform.d_slv_categories.copy()
        self.d_outparms = platform.d_outparms.copy()

        self.clipboard = None
        self.dialog = None

        # Setup the main window
        self.platform = platform
        self.config = platform.config

        log.debug('Application()')
        # Connect all actions to _handle_action
        for x in Actions.get_actions():
            Actions.connect(x, handler=self._handle_action)
            Actions.actions[x].enable()
            if x.startswith('app.'):
                self.add_action(Actions.actions[x])
            # Setup the shortcut keys
            # These are the globally defined shortcuts
            keypress = Actions.actions[x].keypresses
            if keypress:
                self.set_accels_for_action(x, keypress)

        # Initialize
        self.init_file_paths = [os.path.abspath(file_path) for file_path in file_paths]
        self.init = False
        for file_path in file_paths:
            print('Application.py: file_path =', file_path)

    def do_startup(self):
        Gtk.Application.do_startup(self)
        log.debug('Application.do_startup()')

        # Setup the menu
        log.debug('Creating menu')
        '''
        self.menu = Bars.Menu()
        self.set_menu()
        if self.prefers_app_menu():
            self.set_app_menu(self.menu)
        else:
            self.set_menubar(self.menu)
        '''

    def do_activate(self):
        Gtk.Application.do_activate(self)
        log.debug('Application.do_activate()')

        self.main_window = MainWindow(self, self.platform)
        self.main_window.connect('delete-event', self._quit)
        self.get_focus_flag = self.main_window.get_focus_flag

        #setup the messages
        Messages.register_messenger(self.main_window.add_console_line)
        Messages.send_init(self.platform)

        log.debug('Calling Actions.APPLICATION_INITIALIZE')
        Actions.APPLICATION_INITIALIZE()

    def _quit(self, window, event):
        """
        Handle the delete event from the main window.
        Generated by pressing X to close, alt+f4, or right click+close.
        This method in turns calls the state handler to quit.

        Returns:
            true
        """
        Actions.APPLICATION_QUIT()
        return True

    def display_message(self, message):
        dialog_msg = Gtk.MessageDialog(
            transient_for=self.main_window,
            flags=0,
            message_type=Gtk.MessageType.ERROR,
            buttons=Gtk.ButtonsType.OK,
            text=message,
        )
        response = dialog_msg.run()
        if response == Gtk.ResponseType.OK:
            dialog_msg.destroy()

    def show_doc_2(self, pdf_file):
        cmd_exists = lambda x: shutil.which(x) is not None
        l_pdf_readers = ['okular', 'evince', 'xpdf', 'acroread']
        for x in l_pdf_readers:
            if cmd_exists(x):
                cmd = x + ' ' + pdf_file
                os.system(cmd)
                break

    def run_process(self, cmd, s_time):
        process1 = runp.RunProcess(self.main_window, cmd, s_time)

        flag_done = False

        while not flag_done:
            response = process1.run()
            if response == Gtk.ResponseType.OK:
                if psutil.pid_exists(process1.sub_proc.pid):
                    self.display_message('process still running')
                else:

                    if process1.flag_killed:
                        self.display_message('process was stopped by user')
                    elif process1.flag_normal:
                        self.display_message('process successful')
                    elif process1.flag_error:
                        self.display_message('Check for errors in the display window')
                    elif process1.flag_timelimit:
                        self.display_message('time exceeded, process halted')
                    process1.destroy()
                    flag_done = True

            if response == Gtk.ResponseType.CANCEL:

                if psutil.pid_exists(process1.sub_proc.pid):
                    while process1.sub_proc.poll() is None:
                        os.kill(process1.sub_proc.pid, 9)
                    process1.flag_killed = True
                    s_message = "terminated process"
                    self.display_message(s_message)
                else:
                    s_message = "process already completed\nno need to stop"
                    self.display_message(s_message)

    def gparm_add(self, widget, parm_dict):
        dialog = gp.AddGparm(self.main_window)
        flag_done = False

        while not flag_done:
            response = dialog.run()
            if response == Gtk.ResponseType.OK:

                s_name = dialog.entry_name.get_text()
                s_value = dialog.entry_value.get_text()

                if s_name in parm_dict.keys():
                    self.display_message(s_name + ' already exists!')
                elif len(s_name.split()) != 1:
                    self.display_message('gparms name should be a single word.')
                else:
                    parm_dict[s_name] = s_value
                    flag_done = True
                    self.main_window.current_page.saved = False
            elif response == Gtk.ResponseType.CANCEL:
                flag_done = True

        dialog.destroy()

    def gparm_delete(self, widget, parm_dict):
        if len(parm_dict) == 0:
            self.display_message('The gparms list\n is empty!')
            return

        dialog = gp.DelGparm(self.main_window, parm_dict)

        response = dialog.run()
        if response == Gtk.ResponseType.OK:

            l_names = list(parm_dict.keys())
            l_values = list(parm_dict.values())
            parm_dict.clear()

            for i in range(len(l_names)):
                if not dialog.tick[i].get_active():
                    parm_dict[l_names[i]] = l_values[i]

            self.main_window.current_page.saved = False

        dialog.destroy()

    def gparm_edit(self, widget, parm_dict):
        if len(parm_dict) == 0:
            self.display_message('The gparms list\n is empty!')
            return

        dialog = gp.EditGparm(self.main_window, parm_dict)

        response = dialog.run()
        if response == Gtk.ResponseType.OK:

            l_names = list(parm_dict.keys())
            l_values = list(parm_dict.values())

            for i in range(len(l_names)):
                l_values[i] = dialog.value[i].get_text()

            parm_dict.clear()
            for i in range(len(l_names)):
                parm_dict[l_names[i]] = l_values[i]

            self.main_window.current_page.saved = False

        dialog.destroy()

    def get_next_outvar_name(self, ov_dict):
        if len(ov_dict) == 0:
            n_max = -1
        else:
            n_max = max(map(
              lambda x : int(x.split('outvar')[1]) if x.startswith('outvar') else -1 ,
              ov_dict.keys()))
        ovname_next = 'outvar' + str(n_max + 1)
        return ovname_next

    def outvar_add_connection(self, widget, ov_dict, connection):

        ovname_next = self.get_next_outvar_name(ov_dict)

        l1 = list(connection.export_data())

        if l1[0].startswith('ground') or l1[2].startswith('ground'):
            print('outvar_add_connection: ground connection not allowed as outvar.')
            self.display_message('ground connection not allowed as outvar.')
            return
        if l1[1].startswith('b') or l1[3].startswith('b'):
            print('outvar_add_connection: bus connection not allowed as outvar.')
            self.display_message('bus connection not allowed as outvar.')
            return

        for v in ov_dict.values():
            if l1 == v[1]:
                print('outvar_add_connection: wire already exists in outvars.')
                self.display_message('wire already exists in outvars!')
                return

        ov_value = l1

        dialog = ov.AddOutvarConnection(self.main_window, ovname_next, ov_value)
        flag_done = False

        while not flag_done:
            response = dialog.run()
            if response == Gtk.ResponseType.OK:

                s_name = dialog.entry_name.get_text()
                s_value = dialog.entry_value.get_text()

                if s_name in ov_dict.keys():
                    self.display_message(s_name + ' already exists in outvars!')
                elif len(s_name.split()) != 1:
                    self.display_message('outvar name should be a single word.')
                else:
                    ov_dict[s_name] = ['connection', l1]
                    flag_done = True
                    self.main_window.current_page.saved = False
            elif response == Gtk.ResponseType.CANCEL:
                flag_done = True

        dialog.destroy()

    def outvar_add_outvar(self, widget, ov_dict, l_ov_values, block_name):

        if len(l_ov_values) == 0:
            print('outvar_add_outvar: the selected block has no outvars.')
            self.display_message('the selected block has no outvars.')
            return

#       default name for the outvar:
        ovname_next = self.get_next_outvar_name(ov_dict)

        dialog = ov.AddOutvarOutvar(self.main_window, ovname_next, l_ov_values)
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            for i in range(len(l_ov_values)):
                if dialog.l_buttons[i].get_active():
                    i_active = i
                    break
            s_name = dialog.entry_name.get_text()
            s_value = l_ov_values[i_active]
            if s_name in ov_dict.keys():
                self.display_message(s_name + ' already exists in outvars!')
            elif len(s_name.split()) != 1:
                print(s_name, 'outvar name should be a single word.')
                self.display_message('outvar name should be a single word.')
            else:
                l = [block_name, s_value]
                if l in list(map(lambda x: x[1] , ov_dict.values())):
                    self.display_message('outvar already exists!')
                else:
                    ov_dict[s_name] = ['outvar',  [block_name, s_value]]
                    self.main_window.current_page.saved = False

        dialog.destroy()

    def outvar_delete(self, widget, ov_dict, l_output_blocks, l_solve_blocks):
        if len(ov_dict) == 0:
            self.display_message('The outvars list\n is empty!')
            return

        dialog = ov.DelOutvar(self.main_window, ov_dict)

        response = dialog.run()
        if response == Gtk.ResponseType.OK:

            l_names = list(ov_dict.keys())
            l_values = list(ov_dict.values())
            ov_dict.clear()

            for i in range(len(l_names)):
                if not dialog.tick[i].get_active():
                    ov_dict[l_names[i]] = l_values[i]

            for i in range(len(l_names)):
                if dialog.tick[i].get_active():
                    ov_name = l_names[i]
                    for out in l_output_blocks:
                        if ov_name in out.l_outvars:
                            i_ov = out.l_outvars.index(ov_name)
                            del out.l_outvars[i_ov]

            self.main_window.current_page.saved = False

        dialog.destroy()

#       if any output blocks have become empty, remove those

        l_empty_output_blocks = []

        for out in l_output_blocks:
            if len(out.l_outvars) == 0:
                print('outvar_delete: output block <' + out.name + '> is empty')
                l_empty_output_blocks.append(out.name)

        for out_name in l_empty_output_blocks:
            l_out_names = list(map(lambda x: x.name, l_output_blocks))

            i_out = l_out_names.index(out_name)
            i_slv = int(l_output_blocks[i_out].index_slv)

            print('outvar_delete: out_name:', out_name, 'i_out:', i_out, 'i_slv:', i_slv)
            print('type(i_out):', type(i_out), 'type(i_slv):', type(i_slv))

            i_slv_out = l_solve_blocks[i_slv].l_out.index(out_name)
            del l_solve_blocks[i_slv].l_out[i_slv_out]
            del l_output_blocks[i_out]

        self.main_window.current_page.saved = False

    def outvar_edit(self, widget, ov_dict, l_output_blocks):
        if len(ov_dict) == 0:
            self.display_message('The outvars list\n is empty!')
            return

        dialog = ov.EditOutvar(self.main_window, ov_dict)
        flag_done = False

        while not flag_done:
            response = dialog.run()
            if response == Gtk.ResponseType.OK:

                l_names = list(ov_dict.values())
                l_values = list(ov_dict.keys())

                for i in range(len(l_names)):
                    l_values[i] = dialog.value[i].get_text()
                flag_accept = True
                if len(l_values) != len(set(l_values)):
                    self.display_message('Duplicate outvar name found.')
                    flag_accept = False

                if flag_accept:
#                   handle output blocks

                    l_1 = []
                    for out in l_output_blocks:
                        l_2 = []
                        for ov_name in out.l_outvars:
                            l_2.append(ov_dict[ov_name])
                        l_1.append(l_2)

#                   re-assign ov_dict values
                    ov_dict.clear()
                    for i in range(len(l_names)):
                        ov_dict[l_values[i]] = l_names[i]
                    flag_done = True

                    l_new_1 = []
                    for i_out, out in enumerate(l_output_blocks):
                        l_new_2 = []
                        for i_ov, ov_name in enumerate(out.l_outvars):
                            ov_value = l_1[i_out][i_ov]
                            for k, v in ov_dict.items():
                                if ov_value == v:
                                    l_new_2.append(k)
                                    break
                        l_new_1.append(l_new_2)

                    for i_out, out in enumerate(l_output_blocks):
                        for i_ov in range(len(out.l_outvars)):
                            out.l_outvars[i_ov] = l_new_1[i_out][i_ov]

                    self.main_window.current_page.saved = False
            elif response == Gtk.ResponseType.CANCEL:
                flag_done = True

        dialog.destroy()

    def solve_add(self, widget, l_solve_blocks, d_slvparms):
        dialog = slvblk.AddSolveBlock(self.main_window)

        flag_done = False

        while not flag_done:
            response = dialog.run()
            if response == Gtk.ResponseType.OK:

                s_name = dialog.entry_name.get_text()
                s_value = dialog.entry_value.get_text()

                l_slv_names = list(map(lambda x: x.name, l_solve_blocks))
                l_slv_indices = list(map(lambda x: x.index, l_solve_blocks))

                if s_name in l_slv_names:
                    self.display_message(s_name + ' already exists!')
                else:
                    if s_value in l_slv_indices:
                        message='index ' + s_value \
                           + ' is already assigned to another solve block.'
                        self.display_message(message)
                    elif len(s_name.split()) != 1:
                        self.display_message('solve block name should be a single word.')
                    else:
                        slv = SolveBlock(d_slvparms, s_name, s_value)
                        l_solve_blocks.append(slv)
                        l_1 = sorted(l_solve_blocks, key=lambda x: x.index)
                        l_solve_blocks.clear()
                        for slv in l_1:
                            l_solve_blocks.append(slv)
                        flag_done = True
                        self.main_window.current_page.saved = False
            elif response == Gtk.ResponseType.CANCEL:
                flag_done = True

        dialog.destroy()

    def solve_delete(self, widget, l_solve_blocks, l_output_blocks):
        if len(l_solve_blocks) == 0:
            self.display_message('The solve blocks list\n is empty!')
            return

        dialog = slvblk.DelSolveBlock(self.main_window, l_solve_blocks)
        response = dialog.run()

        if response == Gtk.ResponseType.OK:

            l_1 = [slv.name for i, slv in enumerate(l_solve_blocks)
                   if dialog.tick[i].get_active()]

            l_out_names = list(map(lambda x: x.name, l_output_blocks))

            for name1 in l_1:
                for i, slv in enumerate(l_solve_blocks):
                    if slv.name == name1:
                        for out_name in l_solve_blocks[i].l_out:
                            i_out = l_out_names.index(out_name)
                            del l_output_blocks[i_out]

                        del l_solve_blocks[i]
                        break
            self.main_window.current_page.saved = False

        dialog.destroy()

    def solve_pick(self, widget, l_solve_blocks, d_slv_categories):
        if len(l_solve_blocks) == 0:
            self.display_message('The solve blocks list\n is empty!')
            return
        elif len(l_solve_blocks) == 1:
#           some code repetition here for dialog1; clean up later:
            i_active = 0
            dialog1 = slvblk.EditSolveBlock(self.main_window,
                l_solve_blocks[i_active], d_slv_categories)
            response = dialog1.run()
            if response == Gtk.ResponseType.OK:
                gu.assign_parms_1(dialog1.d_widgets_1,
                   l_solve_blocks[i_active].d_parms)
                l_solve_blocks[i_active].index = \
                l_solve_blocks[i_active].d_parms['block_index']
                l_1 = sorted(l_solve_blocks, key=lambda x: x.index)
                l_solve_blocks.clear()
                for slv in l_1:
                    l_solve_blocks.append(slv)

                self.main_window.current_page.saved = False

            dialog1.destroy()
        else:
            dialog = slvblk.PickSolveBlock(self.main_window, l_solve_blocks)

            response = dialog.run()
            if response == Gtk.ResponseType.OK:
#               check which radio button is active and present
#               the corresponding solve block for editing

#               not pythonic: look at it later
                for i in range(len(dialog.l_buttons)):
                    if dialog.l_buttons[i].get_active():
                        i_active = i
                        break

                dialog1 = slvblk.EditSolveBlock(self.main_window,
                    l_solve_blocks[i_active], d_slv_categories)
                response = dialog1.run()
                if response == Gtk.ResponseType.OK:
                    gu.assign_parms_1(dialog1.d_widgets_1,
                       l_solve_blocks[i_active].d_parms)
                    l_solve_blocks[i_active].index = \
                    l_solve_blocks[i_active].d_parms['block_index']
                    l_1 = sorted(l_solve_blocks, key=lambda x: x.index)

                    l_solve_blocks.clear()

                    for slv in l_1:
                        l_solve_blocks.append(slv)

                    self.main_window.current_page.saved = False

                dialog1.destroy()

            dialog.destroy()

    def solve_pick_1(self, widget, l_solve_blocks, d_slv_categories):
        if len(l_solve_blocks) == 0:
            self.display_message('The solve blocks list\n is empty!')
            return
        else:
            if len(l_solve_blocks) == 1:
                i_active = 0
            else:
                dialog = slvblk.PickSolveBlock(self.main_window, l_solve_blocks)
                response = dialog.run()
                if response == Gtk.ResponseType.OK:
                    for i in range(len(dialog.l_buttons)):
                        if dialog.l_buttons[i].get_active():
                            i_active = i
                            break
                dialog.destroy()

            for k, v in self.d_slvparms.items():
                l_solve_blocks[i_active].d_parms[k] = v['default']

    def solve_pick_2(self, widget, l_solve_blocks):
        if len(l_solve_blocks) == 0:
            self.display_message('The solve blocks list\n is empty!')
            k = -1
            return k
        else:
            if len(l_solve_blocks) == 1:
                i_active = 0
            else:
                dialog = slvblk.PickSolveBlock(self.main_window, l_solve_blocks)
                response = dialog.run()
                if response == Gtk.ResponseType.OK:
                    for i in range(len(dialog.l_buttons)):
                        if dialog.l_buttons[i].get_active():
                            i_active = i
                            break
                dialog.destroy()
            return i_active

    def output_add(self, widget, d_outparms, l_output_blocks, l_solve_blocks):
        if len(l_solve_blocks) == 0:
            self.display_message('The solve blocks list\n is empty!')
            return

        dialog = outblk.AddOutputBlock(self.main_window, l_solve_blocks)

        response = dialog.run()
        if response == Gtk.ResponseType.OK:

            s1 = dialog.combo_1.get_active_text()
            print('output_add: s1:', s1)
            slv_name = s1.split(' (')[0]

            l_slv_names = list(map(lambda x: x.name, l_solve_blocks))
            i_slv = l_slv_names.index(slv_name)
            print('output_add: i_slv:', i_slv)

            if len(l_output_blocks) == 0:
                n_max = -1
            else:
                n_max = max(map(lambda x: int(x.name.split('$')[-1]), l_output_blocks))

            out_name = 'out$' + str(n_max+1)
            out_block = OutputBlock(self.d_outparms, str(i_slv), out_name)
            l_output_blocks.append(out_block)
            l_solve_blocks[i_slv].l_out.append(out_name)

            self.main_window.current_page.saved = False

        dialog.destroy()

    def output_delete(self, widget, l_output_blocks, l_solve_blocks):
        if len(l_output_blocks) == 0:
            self.display_message('The output blocks list\n is empty!')
            return

        dialog = outblk.PickOutputBlock(self.main_window, l_output_blocks,
            l_solve_blocks)

        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            for i in range(len(dialog.l_buttons)):
                if dialog.l_buttons[i].get_active():
                    i_active = i
                    break

            l1 = dialog.l_names[i_active].replace(':', ' ').split()
            slv_name, out_name = l1[0], l1[1]

            l_slv_names = list(map(lambda x: x.name, l_solve_blocks))
            l_out_names = list(map(lambda x: x.name, l_output_blocks))

            i_slv = l_slv_names.index(slv_name)
            i_out = l_out_names.index(out_name)

            i_slv_out = l_solve_blocks[i_slv].l_out.index(out_name)
            del l_solve_blocks[i_slv].l_out[i_slv_out]
            del l_output_blocks[i_out]

            self.main_window.current_page.saved = False

        dialog.destroy()

    def output_pick(self, widget, l_output_blocks, l_solve_blocks,
        d_outparms, l_outvars):

        if len(l_output_blocks) == 0:
            self.display_message('The output blocks list\n is empty!')
            return
        elif len(l_output_blocks) == 1:
            i_active = 0
            dialog1 = outblk.EditOutputBlock(self.main_window,
               l_output_blocks[i_active], d_outparms, l_outvars, l_solve_blocks)
            response = dialog1.run()
            if response == Gtk.ResponseType.OK:
                gu.assign_parms_1(dialog1.d_widgets_1,
                   l_output_blocks[i_active].d_parms)

                l_output_blocks[i_active].l_outvars.clear()
                for i, button in enumerate(dialog1.l_ov_buttons):
                    if button.get_active():
                        l_output_blocks[i_active].l_outvars.append(l_outvars[i])

                self.main_window.current_page.saved = False

            dialog1.destroy()
        else:
            dialog = outblk.PickOutputBlock(self.main_window, l_output_blocks,
                l_solve_blocks)

            response = dialog.run()
            if response == Gtk.ResponseType.OK:

                for i in range(len(dialog.l_buttons)):
                    if dialog.l_buttons[i].get_active():
                        i_active = i
                        break
                print('output_pick: i_active:', i_active)

                out_name = dialog.l_names[i_active].split(' ')[-1]
                i_output_block = -1
                for i1, out in enumerate(l_output_blocks):
                    if out.name == out_name:
                        i_output_block = i1
                        break
                if i_output_block == -1:
                    print('output_pick: i_output_block = -1? Halting...')
                    exit(0)

                dialog1 = outblk.EditOutputBlock(self.main_window,
                   l_output_blocks[i_output_block], d_outparms, l_outvars, l_solve_blocks)
                response = dialog1.run()
                if response == Gtk.ResponseType.OK:
                    gu.assign_parms_1(dialog1.d_widgets_1,
                       l_output_blocks[i_active].d_parms)

                    l_output_blocks[i_active].l_outvars.clear()
                    for i, button in enumerate(dialog1.l_ov_buttons):
                        if button.get_active():
                            l_output_blocks[i_active].l_outvars.append(l_outvars[i])

                    self.main_window.current_page.saved = False

                dialog1.destroy()

            dialog.destroy()

    def element_display(self, widget, e1, ov_dict):

        l1 = []
        for k, v in ov_dict.items():
            l2 = []
            s1 = v[0]
            if isinstance(e1, Connection):
                if s1 == 'connection':
                    if v[1] == list(e1.export_data()):
                        l2.append(k)
                        l2.append(str(v[1]))
            else:
                if s1 == 'outvar':
                    if v[1][0] == e1.name:
                        l2.append(k)
                        l2.append(str(v[1]))

            if l2:
                l1.append(l2)

        if isinstance(e1, Connection):
            s_name = str(list(e1.export_data()))
        else:
            s_name = e1.name

        dialog = disp.DisplayName(self.main_window, s_name, l1)
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            dialog.destroy()

        dialog.destroy()

    def show_param_select(self, widget, selected_block):
        l_drop = [
          'port_sep_x',
          'port_sep_y',
          'port_block_x',
          'port_block_y',
          'port_offset_l',
          'port_offset_r',
          'port_offset_t',
          'port_offset_b',
          'rotate_strict',
          'drawing_scheme',
          'mirror',
          'comment',
        ]
        l_parm_names_1 = list(selected_block.params.keys())
        l_parm_names = [x for x in l_parm_names_1 if x not in l_drop]

        if len(l_parm_names) == 0:
            print('show_param_select: the selected block has no parameters.')
            self.display_message('the selected block has no parameters.')
            return

        dialog = sparm.AddShowParm(self.main_window, l_parm_names)
        response = dialog.run()
        if response == Gtk.ResponseType.OK:
            for i in range(len(l_parm_names)):
                if dialog.l_buttons[i].get_active():
                    i_active = i
                    break
            parm_name = l_parm_names[i_active]

            main = self.main_window
            page = main.current_page
            flow_graph = page.flow_graph if page else None

            x, y = selected_block.coordinate
            if selected_block.rotation in (0, 180):
                offset1 = selected_block.height
            else:
                offset1 = selected_block.width
            c_new = (x, y + offset1 + Constants.SHOW_PARAMETER_SEP)
            b_new_name = flow_graph.add_new_block('show_parameter', c_new)

            b_new = [x for x in flow_graph.blocks if x.name == b_new_name][0]

            b_new.params['element_name'].set_value(selected_block.name)
            b_new.params['parameter_name'].set_value(parm_name)

            flow_graph.update()

        dialog.destroy()

    def _handle_action(self, action, *args):

        log.debug('_handle_action({0}, {1})'.format(action, args))
        main = self.main_window
        page = main.current_page
        flow_graph = page.flow_graph if page else None

        def flow_graph_update(fg=flow_graph):
            main.vars.update_gui(fg.blocks)
            fg.update()

        # Initialize/Quit
        if action == Actions.APPLICATION_INITIALIZE:
            log.debug('APPLICATION_INITIALIZE')
            file_path_to_show = self.config.file_open()
            for file_path in (self.init_file_paths or self.config.get_open_files()):
                if os.path.exists(file_path):
                    main.new_page(file_path, show=file_path_to_show == file_path)
            if not main.current_page:
                main.new_page()  # ensure that at least a blank page exists

            main.btwin.search_entry.hide()

            for action in (
                Actions.ELEMENT_DELETE,
                Actions.BLOCK_PARAM_MODIFY,
                Actions.BLOCK_ROTATE_CCW,
                Actions.BLOCK_ROTATE_CW,
                Actions.BLOCK_VALIGN_TOP,
                Actions.BLOCK_VALIGN_MIDDLE,
                Actions.BLOCK_VALIGN_BOTTOM,
                Actions.BLOCK_HALIGN_LEFT,
                Actions.BLOCK_HALIGN_CENTER,
                Actions.BLOCK_HALIGN_RIGHT,
                Actions.BLOCK_CUT,
                Actions.BLOCK_COPY,
                Actions.BLOCK_PASTE,
                Actions.OPEN_HIER,
                Actions.FLOW_GRAPH_SAVE,
                Actions.FLOW_GRAPH_UNDO,
                Actions.FLOW_GRAPH_REDO,
            ):
                action.disable()

            # Load preferences
            for action in (
                Actions.TOGGLE_BLOCKS_WINDOW,
                Actions.TOGGLE_CONSOLE_WINDOW,
                Actions.TOGGLE_SCROLL_LOCK,
                Actions.TOGGLE_AUTO_HIDE_PORT_LABELS,
                Actions.TOGGLE_SNAP_TO_GRID,
            ):
                action.set_enabled(True)
                if hasattr(action, 'load_from_preferences'):
                    action.load_from_preferences()

            # Hide the panels *IF* it's saved in preferences
            main.update_panel_visibility(main.BLOCKS, Actions.TOGGLE_BLOCKS_WINDOW.get_active())
            main.update_panel_visibility(main.CONSOLE, Actions.TOGGLE_CONSOLE_WINDOW.get_active())

            # Force an update on the current page to match loaded preferences.
            # In the future, change the __init__ order to load preferences first
            page = main.current_page
            if page:
                page.flow_graph.update()

            self.init = True
        elif action == Actions.APPLICATION_QUIT:
            if main.close_pages():
                while Gtk.main_level():
                    Gtk.main_quit()
                exit(0)
        # Selections
        elif action == Actions.ELEMENT_SELECT:
            pass #do nothing, update routines below
        elif action == Actions.NOTHING_SELECT:
            flow_graph.unselect()
        elif action == Actions.SELECT_ALL:
            flow_graph.select_all()
        # Enable/Disable
        # Cut/Copy/Paste
        elif action == Actions.BLOCK_CUT:
            Actions.BLOCK_COPY()
            Actions.ELEMENT_DELETE()
        elif action == Actions.BLOCK_COPY:
            self.clipboard = flow_graph.copy_to_clipboard()
        elif action == Actions.BLOCK_PASTE:
            if self.clipboard:
                flow_graph.paste_from_clipboard(self.clipboard)
                flow_graph_update()
                page.state_cache.save_new_state(flow_graph.export_data())
                page.saved = False
        # Create hier block

        elif action == Actions.GPARM_ADD:

            self.gparm_add(self.main_window, flow_graph.gparms)

        elif action == Actions.GPARM_DEL:
            self.gparm_delete(self.main_window, flow_graph.gparms)
        elif action == Actions.GPARM_EDIT:
            if 'hb' in flow_graph.get_option('generate_options'):
                if 'drawing_scheme' not in flow_graph.gparms.keys():
                    flow_graph.gparms['drawing_scheme'] = 'name'
                if 'rotate_strict' not in flow_graph.gparms.keys():
                    flow_graph.gparms['rotate_strict'] = 'no'
                if 'mirror' not in flow_graph.gparms.keys():
                    flow_graph.gparms['mirror'] = 'none'
                if 'port_sep_x' not in flow_graph.gparms.keys():
                    flow_graph.gparms['port_sep_x'] = '4'
                if 'port_sep_y' not in flow_graph.gparms.keys():
                    flow_graph.gparms['port_sep_y'] = '4'
                if 'port_block_x' not in flow_graph.gparms.keys():
                    flow_graph.gparms['port_block_x'] = '4'
                if 'port_block_y' not in flow_graph.gparms.keys():
                    flow_graph.gparms['port_block_y'] = '4'
                if 'port_offset_l' not in flow_graph.gparms.keys():
                    flow_graph.gparms['port_offset_l'] = '0'
                if 'port_offset_r' not in flow_graph.gparms.keys():
                    flow_graph.gparms['port_offset_r'] = '0'
                if 'port_offset_t' not in flow_graph.gparms.keys():
                    flow_graph.gparms['port_offset_t'] = '0'
                if 'port_offset_b' not in flow_graph.gparms.keys():
                    flow_graph.gparms['port_offset_b'] = '0'

            self.gparm_edit(self.main_window, flow_graph.gparms)
        elif action == Actions.OUTVAR_ADD:
            n1 = len(flow_graph.selected_elements)

            if n1 == 1:
                e1 = list(flow_graph.selected_elements)[0]
                if isinstance(e1, Connection):
                    self.outvar_add_connection(self.main_window, flow_graph.outvars, e1)
                else:
                    k = str(type(e1)).split("'")[-2].split('.')[-1]
                    l_outvars = self.dict_outvars[k]
                    self.outvar_add_outvar(self.main_window,
                        flow_graph.outvars, l_outvars, e1.name)

        elif action == Actions.OUTVAR_DEL:
            self.outvar_delete(self.main_window, flow_graph.outvars,
               flow_graph.l_output_blocks, flow_graph.l_solve_blocks)
        elif action == Actions.OUTVAR_EDIT:
            self.outvar_edit(self.main_window, flow_graph.outvars,
               flow_graph.l_output_blocks)
        elif action == Actions.SOLVEBLOCK_ADD:
            self.solve_add(self.main_window, flow_graph.l_solve_blocks,
               self.d_slvparms)
        elif action == Actions.SOLVEBLOCK_DEL:
            self.solve_delete(self.main_window, flow_graph.l_solve_blocks,
            flow_graph.l_output_blocks)
        elif action == Actions.SOLVEBLOCK_EDIT:
            self.solve_pick(self.main_window, flow_graph.l_solve_blocks,
               self.d_slv_categories)
        elif action == Actions.SOLVEBLOCK_RESET:
            self.solve_pick_1(self.main_window, flow_graph.l_solve_blocks,
               self.d_slv_categories)
        elif action == Actions.SOLVEBLOCK_DISP:
            print('solve disp clicked')
            i_picked = self.solve_pick_2(self.main_window, flow_graph.l_solve_blocks)
            print('i_picked:', i_picked)
            if i_picked > -1:

                l_out_names = list(map(lambda x: x.name, flow_graph.l_output_blocks))

                slv = flow_graph.l_solve_blocks[i_picked]
                s_title = 'Solve Block: ' + slv.name + ' (' + slv.index + ')'
                print('s_title:', s_title)
                l = []
                for k, v in slv.d_parms.items(): l.append([k, v])

                n1 = len(slv.l_out)
                if n1 != 0:
                    s1 = 'output blocks'
                    s2 = ''
                    for i, out_name in enumerate(slv.l_out):
                        s2 += out_name
                        if i < (n1-1): s2 += ', '
                    l.append([s1, s2])

                for out_name in slv.l_out:
                    i_out = l_out_names.index(out_name)
                    out = flow_graph.l_output_blocks[i_out]
                    for k, v in out.d_parms.items():
                        l.append([out_name + ': ' + k, v])
                    n1 = len(out.l_outvars)
                    s1 = out_name + ': outvars:'
                    s2 = ''
                    for i_ov, ov_name in enumerate(out.l_outvars):
                        s2 += ov_name
                        if i_ov < (n1-1): s2 += ', '
                    l.append([s1, s2])

                nmax = 0
                for l1 in l:
                    n2 = len(l1[0])
                    if n2 > nmax: nmax = n2
                    l1[0] += ' '
                print('nmax:', nmax)
                nmax1 = nmax + 7
                for l1 in l:
                    l1[0] = ' ' + l1[0].ljust(nmax1, '-') + ' '

                for l1 in l: print(l1)

                window_W = 450
                window_H = 600

                slist1 = slist.ShowList(l, window_W, window_H, s_title)
                slist1.show_all()

        elif action == Actions.OUTPUTBLOCK_ADD:
            self.output_add(self.main_window, self.d_outparms,
               flow_graph.l_output_blocks, flow_graph.l_solve_blocks)
        elif action == Actions.OUTPUTBLOCK_DEL:
            self.output_delete(self.main_window,
               flow_graph.l_output_blocks, flow_graph.l_solve_blocks)
        elif action == Actions.OUTPUTBLOCK_EDIT:

            l_outvars_1 = list(flow_graph.outvars.keys())
            self.output_pick(self.main_window,
               flow_graph.l_output_blocks, flow_graph.l_solve_blocks,
               self.d_outparms, l_outvars_1)
        elif action == Actions.ELEMENT_DISPLAY:
            n1 = len(flow_graph.selected_elements)

            if n1 == 1:
                e1 = list(flow_graph.selected_elements)[0]
                self.element_display(self.main_window, e1, flow_graph.outvars)

        elif action == Actions.DOC_DISPLAY:
            e1 = list(flow_graph.selected_elements)[0]
            show_doc_1(self.platform, e1.key)

        elif action == Actions.SHOW_PARAM:
            print('Application.py: show parameter selected.')

            n1 = len(flow_graph.selected_elements)
            print('paste selected clicked: n1:', n1)
            if n1 == 1:
                e1 = list(flow_graph.selected_elements)[0]
                print('outvar add clicked: type(e1):', type(e1))
                self.show_param_select(self.main_window, e1)

        elif action == Actions.PASTE_SELECTED:
            n1 = len(flow_graph.selected_elements)
            print('paste selected clicked: n1:', n1)
            x_off = flow_graph.click_right_pos_x
            y_off = flow_graph.click_right_pos_y
            print('flow_graph.click_right_pos_x, y:', x_off, y_off)
            if self.clipboard:
                flow_graph.paste_from_clipboard_1(self.clipboard, x_off, y_off)
                flow_graph_update()
                page.state_cache.save_new_state(flow_graph.export_data())
                page.saved = False

        # Move/Rotate/Delete/Create
        elif action == Actions.BLOCK_MOVE:
            page.state_cache.save_new_state(flow_graph.export_data())
            page.saved = False
        elif action in Actions.BLOCK_ALIGNMENTS:
            if flow_graph.align_selected(action):
                page.state_cache.save_new_state(flow_graph.export_data())
                page.saved = False
        elif action == Actions.BLOCK_ROTATE_CCW:
            if flow_graph.rotate_selected(90):
                flow_graph_update()
                page.state_cache.save_new_state(flow_graph.export_data())
                page.saved = False
        elif action == Actions.BLOCK_ROTATE_CW:
            if flow_graph.rotate_selected(-90):
                flow_graph_update()
                page.state_cache.save_new_state(flow_graph.export_data())
                page.saved = False
        elif action == Actions.ELEMENT_DELETE:
            if flow_graph.remove_selected():
                flow_graph_update()
                page.state_cache.save_new_state(flow_graph.export_data())
                Actions.NOTHING_SELECT()
                page.saved = False
        elif action == Actions.ELEMENT_CREATE:
            flow_graph_update()
            page.state_cache.save_new_state(flow_graph.export_data())
            Actions.NOTHING_SELECT()
            page.saved = False
        # Window stuff
        elif action == Actions.ABOUT_WINDOW_DISPLAY:
            Dialogs.show_about(main, self.platform.config)
        elif action == Actions.HELP_WINDOW_DISPLAY:
            Dialogs.show_help(main)
        elif action == Actions.KEYBOARD_SHORTCUTS_WINDOW_DISPLAY:
            Dialogs.show_keyboard_shortcuts(main)   
        elif action == Actions.SHOWDOC_WINDOW_DISPLAY:

            l3 = page.file_path.rsplit('/', 1)
            pdf_file = l3[0] + '/proj_' + l3[1].replace('.grc', '.pdf')

            if (os.path.exists(pdf_file)):
                self.show_doc_2(pdf_file)
            else:
                s = pdf_file + ' was not found.'
                self.display_message(s)
        elif action == Actions.TOGGLE_CONSOLE_WINDOW:
            action.set_active(not action.get_active())
            main.update_panel_visibility(main.CONSOLE, action.get_active())
            action.save_to_preferences()
        elif action == Actions.TOGGLE_BLOCKS_WINDOW:
            # This would be better matched to a Gio.PropertyAction, but to do
            # this, actions would have to be defined in the window not globally
            action.set_active(not action.get_active())
            main.update_panel_visibility(main.BLOCKS, action.get_active())
            action.save_to_preferences()
        elif action == Actions.TOGGLE_SCROLL_LOCK:
            action.set_active(not action.get_active())
            active = action.get_active()
            main.console.text_display.scroll_lock = active
            if active:
                main.console.text_display.scroll_to_end()
            action.save_to_preferences()
        elif action == Actions.CLEAR_CONSOLE:
            main.console.text_display.clear()
        elif action == Actions.SAVE_CONSOLE:
            file_path = FileDialogs.SaveConsole(main, page.file_path).run()
            if file_path is not None:
                main.console.text_display.save(file_path)
        elif action == Actions.TOGGLE_AUTO_HIDE_PORT_LABELS:
            action.set_active(not action.get_active())
            action.save_to_preferences()
            for page in main.get_pages():
                page.flow_graph.create_shapes()
        elif action == Actions.TOGGLE_SNAP_TO_GRID:
            action.set_active(not action.get_active())
            action.save_to_preferences()
        # Param Modifications
        elif action == Actions.BLOCK_PARAM_MODIFY:
            selected_block = args[0] if args[0] else flow_graph.selected_block
            if selected_block:
                self.dialog = PropsDialog(self.main_window, selected_block)
                response = Gtk.ResponseType.APPLY
                while response == Gtk.ResponseType.APPLY:  # rerun the dialog if Apply was hit
                    response = self.dialog.run()
                    if response in (Gtk.ResponseType.APPLY, Gtk.ResponseType.ACCEPT):
                        page.state_cache.save_new_state(flow_graph.export_data())
                        n = page.state_cache.get_current_state()
                        flow_graph.import_data(n)
                        flow_graph_update()

                        page.saved = False
                    else:  # restore the current state
                        n = page.state_cache.get_current_state()
                        flow_graph.import_data(n)
                        flow_graph_update()
                    if response == Gtk.ResponseType.APPLY:
                        # null action, that updates the main window
                        Actions.ELEMENT_SELECT()
                self.dialog.destroy()
                self.dialog = None
        elif action == Actions.EXTERNAL_UPDATE:
            page.state_cache.save_new_state(flow_graph.export_data())
            flow_graph_update()
            if self.dialog is not None:
                self.dialog.update_gui(force=True)
            page.saved = False
        elif action == Actions.VARIABLE_EDITOR_UPDATE:
            page.state_cache.save_new_state(flow_graph.export_data())
            flow_graph_update()
            page.saved = False
        # View Parser Errors
        # Undo/Redo
        elif action == Actions.FLOW_GRAPH_UNDO:
            n = page.state_cache.get_prev_state()
            if n:
                flow_graph.unselect()
                flow_graph.import_data(n)
                flow_graph_update()
                page.saved = False
        elif action == Actions.FLOW_GRAPH_REDO:
            n = page.state_cache.get_next_state()
            if n:
                flow_graph.unselect()
                flow_graph.import_data(n)
                flow_graph_update()
                page.saved = False
        # New/Open/Save/Close
        elif action == Actions.FLOW_GRAPH_NEW:
            main.new_page()
            args = (GLib.Variant('s', 'qt_gui'),)
            flow_graph = main.current_page.flow_graph
            flow_graph.options_block.params['generate_options'].set_value(str(args[0])[1:-1])
            flow_graph.options_block.params['author'].set_value(getuser())
            flow_graph_update(flow_graph)
        elif action == Actions.FLOW_GRAPH_NEW_TYPE:
            main.new_page()
            if args:
                flow_graph = main.current_page.flow_graph
                flow_graph.options_block.params['generate_options'].set_value(str(args[0])[1:-1])
                flow_graph_update(flow_graph)
        elif action == Actions.FLOW_GRAPH_OPEN:
            file_paths = args[0] if args[0] else FileDialogs.OpenFlowGraph(main, page.file_path).run()
            if file_paths: # Open a new page for each file, show only the first
                for i,file_path in enumerate(file_paths):
                    main.new_page(file_path, show=(i==0))
                    self.config.add_recent_file(file_path)
                    main.tool_bar.refresh_submenus()
                    #main.menu_bar.refresh_submenus()
        elif action == Actions.FLOW_GRAPH_CLOSE:
            main.close_page()
        elif action == Actions.FLOW_GRAPH_OPEN_RECENT:
            file_path = str(args[0])[1:-1]
            main.new_page(file_path, show=True)
            main.tool_bar.refresh_submenus()
        elif action == Actions.FLOW_GRAPH_SAVE:
            #read-only or undefined file path, do save-as
            if page.get_read_only() or not page.file_path:
                Actions.FLOW_GRAPH_SAVE_AS()
            #otherwise try to save
            else:
                try:
                    self.platform.save_flow_graph(page.file_path, flow_graph)
                    flow_graph.grc_file_path = page.file_path
                    page.saved = True
                except IOError:
                    Messages.send_fail_save(page.file_path)
                    page.saved = False
        elif action == Actions.FLOW_GRAPH_SAVE_AS:
            file_path = FileDialogs.SaveFlowGraph(main, page.file_path).run()

            if file_path is not None:
                if flow_graph.options_block.params['id'].get_value() == 'default':
                    file_name = os.path.basename(file_path).replace('.grc', '')
                    flow_graph.options_block.params['id'].set_value(file_name)
                    flow_graph_update(flow_graph)

                page.file_path = os.path.abspath(file_path)
                try:
                    self.platform.save_flow_graph(page.file_path, flow_graph)
                    flow_graph.grc_file_path = page.file_path
                    page.saved = True
                except IOError:
                    Messages.send_fail_save(page.file_path)
                    page.saved = False
                self.config.add_recent_file(file_path)
                main.tool_bar.refresh_submenus()
                #TODO
                #main.menu_bar.refresh_submenus()
        elif action == Actions.FLOW_GRAPH_SAVE_COPY:
            try:
                if not page.file_path:
                    # Make sure the current flowgraph has been saved
                    Actions.FLOW_GRAPH_SAVE_AS()
                else:
                    dup_file_path = page.file_path
                    dup_file_name = '.'.join(dup_file_path.split('.')[:-1]) + '_copy' # Assuming .grc extension at the end of file_path
                    dup_file_path_temp = dup_file_name + Constants.FILE_EXTENSION
                    count = 1
                    while os.path.exists(dup_file_path_temp):
                        dup_file_path_temp = '{}({}){}'.format(dup_file_name, count, Constants.FILE_EXTENSION)
                        count += 1
                    dup_file_path_user = FileDialogs.SaveFlowGraph(main, dup_file_path_temp).run()
                    if dup_file_path_user is not None:
                        self.platform.save_flow_graph(dup_file_path_user, flow_graph)
                        Messages.send('Saved Copy to: "' + dup_file_path_user + '"\n')
            except IOError:
                Messages.send_fail_save('Can not create a copy of the flowgraph\n')
        elif action == Actions.FLOW_GRAPH_DUPLICATE:
            previous = flow_graph
            # Create a new page
            main.new_page()
            page = main.current_page
            new_flow_graph = page.flow_graph
            # Import the old data and mark the current as not saved
            new_flow_graph.import_data(previous.export_data())
            flow_graph_update(new_flow_graph)
            page.state_cache.save_new_state(new_flow_graph.export_data())
            page.saved = False
        elif action == Actions.FLOW_GRAPH_SCREEN_CAPTURE:
            file_path, background_transparent = FileDialogs.SaveScreenShot(main, page.file_path).run()
            if file_path is not None:
                try:
                    Utils.make_screenshot(flow_graph, file_path, background_transparent)
                except ValueError:
                    Messages.send('Failed to generate screen shot\n')
        elif action == Actions.FLOW_GRAPH_GEN:
            if not page.saved or not page.file_path:
                Actions.FLOW_GRAPH_SAVE()  # only save if file path missing or not saved
            if page.saved and page.file_path:
                flow_graph = self.main_window.current_page.flow_graph
                if 'hb' in flow_graph.get_option('generate_options'):
                    generator = page.get_generator()
                    try:
                        Messages.send_start_gen(generator.file_path)
                        generator.write(flow_graph)
                    except Exception as e:
                        Messages.send_fail_gen(e)
                else:
                    cctfile_name = self.config.gseim_output_dir + '/' \
                       + page.file_path.split('/')[-1].split('.')[0] + '.in'

                    cmd = 'python3'
                    cmd += ' ' + self.config.gseim_exec_dir + '/gseim1.py'
#                   cmd += ' ' + self.config.gseim_output_dir + '/cct.in'
                    cmd += ' ' + cctfile_name
                    cmd += ' ' + self.config.block_lib_dir + '/'
                    cmd += ' ' + self.config.hier_block_lib_dir + '/'
                    cmd += ' ' + self.config.gseim_xbe_dir + '/'
                    cmd += ' ' + self.config.gseim_ebe_dir + '/'
                    cmd += ' ' + self.config.gseim_bbe_dir + '/'
                    cmd += ' ' + page.file_path

                    print('grc/gui/Application.py: cmd:', cmd)

                    self.run_process(cmd, '00:10:00')

        elif action == Actions.FLOW_GRAPH_EXEC:
            flow_graph = self.main_window.current_page.flow_graph

            if 'hb' in flow_graph.get_option('generate_options'):
                self.display_message('cannot run sub-circuit')
            else:
#               check if the flowgraph needs to be saved
                if not page.saved or not page.file_path:
                    Actions.FLOW_GRAPH_SAVE()
#               check if the .grc file is newer than the .in file

                cctfile_name = self.config.gseim_output_dir + '/' \
                  + page.file_path.split('/')[-1].split('.')[0] + '.in'

                if os.path.exists(cctfile_name):
                    t1 = os.path.getmtime(cctfile_name)
                    t2 = os.path.getmtime(page.file_path)
                    if (t2 > t1):
                        Actions.FLOW_GRAPH_GEN()
                else:
                    Actions.FLOW_GRAPH_GEN()

                cmd = self.config.gseim_cpp_dir + '/gsmain'
                cmd += ' ' + cctfile_name

                print('grc/gui/Application.py: cmd:', cmd)
                s = flow_graph.options_block.params['engine_output'].get_value()
                if s == 'GUI_window':
                    self.run_process(cmd, '00:10:00')
                elif s == 'console_window':
                    message = 'Engine output sent to console.\n'\
                       + '"GSEIM: Program Completed" will be displayed'\
                       + ' if it runs successfully.\n'\
                       + 'To view engine output in a GUI window, set the "engine output"'\
                       + ' parameter of the Options block to "GUI Window".'
                    self.display_message(message)
                    subprocess.Popen(cmd, shell=True)

        elif action == Actions.FLOW_GRAPH_PLOT:
            cmd = 'python3 ' + self.config.gseim_exec_dir + '/gseim_plot.py'
            cmd += ' ' + self.config.gseim_output_dir + '/'
            cmd += ' ' + self.config.gseim_output_dir + '/gseim_plot_history.dat'

            print('grc/gui/Application.py: cmd:', cmd)
            os.system(cmd)

        elif action == Actions.PAGE_CHANGE:  # pass and run the global actions
            flow_graph_update()
        elif action == Actions.FIND_BLOCKS:
            flow_graph.unselect()
            main.update_panel_visibility(main.BLOCKS, True)
            main.btwin.search_entry.show()
            main.btwin.search_entry.grab_focus()
        elif action == Actions.OPEN_HIER:

            if len(list(flow_graph.selected_blocks())) == 1:
                for b in flow_graph.selected_blocks():
                    grc_source = b.extra_data.get('grc_source', '')
                    if grc_source:
                        main.new_page(grc_source, show=True)
                    else:
                        self.display_message("open hier allowed only for hier blocks")
            else:
                self.display_message("open hier allowed only when a single block is selected")

        else:
            log.warning('!!! Action "%s" not handled !!!' % action)
        # Global Actions for all States
        page = main.current_page  # page and flow graph might have changed
        flow_graph = page.flow_graph if page else None

        selected_blocks = list(flow_graph.selected_blocks())
        selected_block = selected_blocks[0] if selected_blocks else None

        #update general buttons
        Actions.ELEMENT_DELETE.set_enabled(bool(flow_graph.selected_elements))
        Actions.BLOCK_PARAM_MODIFY.set_enabled(bool(selected_block))
        Actions.BLOCK_ROTATE_CCW.set_enabled(bool(selected_blocks))
        Actions.BLOCK_ROTATE_CW.set_enabled(bool(selected_blocks))
        #update alignment options
        for act in Actions.BLOCK_ALIGNMENTS:
            if act:
                act.set_enabled(len(selected_blocks) > 1)
        #update cut/copy/paste
        Actions.BLOCK_CUT.set_enabled(bool(selected_blocks))
        Actions.BLOCK_COPY.set_enabled(bool(selected_blocks))
        Actions.BLOCK_PASTE.set_enabled(bool(self.clipboard))

        Actions.OPEN_HIER.set_enabled(bool(selected_blocks))

        Actions.FIND_BLOCKS.enable()

        Actions.FLOW_GRAPH_SAVE.set_enabled(not page.saved)
        main.update()

        flow_graph.update_selected()
        page.drawing_area.queue_draw()

        return True  # Action was handled

